# Copyright 2021 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose

from __future__ import annotations

import json
import logging
import os
from observatory.platform.elastic.elastic import Elastic
from observatory.platform.elastic.kibana import Kibana
from observatory.platform.utils.config_utils import module_file_path
from observatory.platform.utils.file_utils import load_file
from observatory.platform.utils.jinja2_utils import render_template
from observatory.platform.utils.test_utils import ObservatoryEnvironment, ObservatoryTestCase
from observatory.platform.utils.workflow_utils import make_dag_id
from observatory.platform.workflows.elastic_import_workflow import load_elastic_mappings_simple
from typing import Dict

from academic_observatory_workflows.config import elastic_mappings_folder
from academic_observatory_workflows.dags.elastic_import_workflow import load_elastic_mappings_ao


class TestElasticImportWorkflow(ObservatoryTestCase):
    """Tests for the Elastic Import Workflow"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")

    def test_load_elastic_mappings_ao(self):
        """Test load_elastic_mappings_ao"""

        path = elastic_mappings_folder()
        aggregate = "author"
        expected = [
            ("ao_dois", load_file(os.path.join(path, "ao-dois-mappings.json"))),
            (
                "ao_author_access_types",
                render_template(
                    os.path.join(path, "ao-access-types-mappings.json.jinja2"),
                    aggregate=aggregate,
                    facet="access_types",
                ),
            ),
            (
                "ao_author_disciplines",
                render_template(
                    os.path.join(path, "ao-disciplines-mappings.json.jinja2"), aggregate=aggregate, facet="disciplines"
                ),
            ),
            (
                "ao_author_events",
                render_template(
                    os.path.join(path, "ao-events-mappings.json.jinja2"), aggregate=aggregate, facet="events"
                ),
            ),
            (
                "ao_author_metrics",
                render_template(
                    os.path.join(path, "ao-metrics-mappings.json.jinja2"), aggregate=aggregate, facet="metrics"
                ),
            ),
            (
                "ao_author_output_types",
                render_template(
                    os.path.join(path, "ao-output-types-mappings.json.jinja2"),
                    aggregate=aggregate,
                    facet="output_types",
                ),
            ),
            (
                "ao_author_unique_list",
                render_template(
                    os.path.join(path, "ao-unique-list-mappings.json.jinja2"), aggregate=aggregate, facet="unique_list"
                ),
            ),
            (
                "ao_author_output_types",
                render_template(
                    os.path.join(path, "ao-output-types-mappings.json.jinja2"),
                    aggregate=aggregate,
                    facet="output_types",
                ),
            ),
            (
                "ao_author_countries",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="countries"
                ),
            ),
            (
                "ao_author_funders",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="funders"
                ),
            ),
            (
                "ao_author_groupings",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="groupings"
                ),
            ),
            (
                "ao_author_institutions",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="institutions"
                ),
            ),
            (
                "ao_author_journals",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="journals"
                ),
            ),
            (
                "ao_author_publishers",
                render_template(
                    os.path.join(path, "ao-relations-mappings.json.jinja2"), aggregate=aggregate, facet="publishers"
                ),
            ),
        ]

        for table_id, expected_mappings_str in expected:
            logging.info(table_id)
            expected_mappings = json.loads(expected_mappings_str)
            actual_mappings = load_elastic_mappings_ao(path, table_id)
            self.assertEqual(expected_mappings, actual_mappings)

    def test_dag_load(self):
        """Test that the DAG can be loaded from a DAG bag.

        :return: None
        """

        env = ObservatoryEnvironment(self.project_id, self.data_location, enable_api=False)
        with env.create():
            expected_dag_ids = [make_dag_id("elastic_import", suffix) for suffix in ["observatory"]]

            dag_file = os.path.join(
                module_file_path("academic_observatory_workflows.dags"), "elastic_import_workflow.py"
            )
            for dag_id in expected_dag_ids:
                self.assert_dag_load(dag_id, dag_file)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

import gzip
import io
import os
from datetime import timedelta
from types import SimpleNamespace
from unittest.mock import ANY, patch

import pendulum
from airflow.exceptions import AirflowException, AirflowSkipException
from airflow.models.connection import Connection
from airflow.models.variable import Variable
from botocore.response import StreamingBody
from click.testing import CliRunner
from observatory.platform.utils.airflow_utils import AirflowConns, AirflowVars, BaseHook
from observatory.platform.utils.gc_utils import (
    upload_file_to_cloud_storage,
    upload_files_to_cloud_storage,
)
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.workflow_utils import blob_name

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.orcid_telescope import OrcidRelease, OrcidTelescope, transform_single_file


class TestOrcidTelescope(ObservatoryTestCase):
    """Tests for the ORCID telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.
        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super(TestOrcidTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")

        self.first_execution_date = pendulum.datetime(year=2021, month=2, day=1)
        self.second_execution_date = pendulum.datetime(year=2021, month=3, day=1)

        # orcid records
        self.records = {}
        for file in ["0000-0002-9227-8610.xml", "0000-0002-9228-8514.xml", "0000-0002-9229-8514.xml"]:
            self.records[file] = {
                "blob": f"{file[-7:-4]}/{file}",
                "path": test_fixtures_folder("orcid", file),
            }
        # last modified file
        self.last_modified_path = test_fixtures_folder("orcid", "last_modified.csv.tar")

        # release used for method tests
        self.release = OrcidRelease(
            OrcidTelescope.DAG_ID, pendulum.datetime(2020, 1, 1), pendulum.datetime(2020, 2, 1), False, max_processes=1
        )

    def test_dag_structure(self):
        """Test that the ORCID DAG has the correct structure.
        :return: None
        """

        dag = OrcidTelescope().make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["get_release_info"],
                "get_release_info": ["transfer"],
                "transfer": ["download_transferred"],
                "download_transferred": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load_partition"],
                "bq_load_partition": ["bq_delete_old"],
                "bq_delete_old": ["bq_append_new"],
                "bq_append_new": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the ORCID DAG can be loaded from a DAG bag.
        :return: None
        """

        with ObservatoryEnvironment().create():
            dag_file = os.path.join(module_file_path("academic_observatory_workflows.dags"), "orcid_telescope.py")
            self.assert_dag_load("orcid", dag_file)

    @patch("academic_observatory_workflows.workflows.orcid_telescope.aws_to_google_cloud_storage_transfer")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.boto3.client")
    def test_telescope(self, mock_client, mock_transfer):
        """Test the ORCID telescope end to end.
        :return: None.
        """
        # Setup Observatory environment
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        dataset_id = env.add_dataset()

        # Set up google cloud sync bucket
        orcid_bucket = env.add_bucket()

        # Setup Telescope
        telescope = OrcidTelescope(dataset_id=dataset_id)
        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create():
            # Add connection
            conn = Connection(conn_id=AirflowConns.ORCID, uri="aws://UWLA41aAhdja:AJLD91saAJSKAL0AjAhkaka@")
            # uri=self.orcid_conn)
            env.add_connection(conn)

            # Add variable
            var = Variable(key=AirflowVars.ORCID_BUCKET, val=orcid_bucket)  # type: ignore
            env.add_variable(var)

            # first run
            with env.create_dag_run(dag, self.first_execution_date):
                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__)

                # Test list releases task with files available
                ti = env.run_task(telescope.get_release_info.__name__)
                start_date, end_date, first_release = ti.xcom_pull(
                    key=OrcidTelescope.RELEASE_INFO,
                    task_ids=telescope.get_release_info.__name__,
                    include_prior_dates=False,
                )
                self.assertEqual(pendulum.parse(start_date), dag.default_args["start_date"])
                self.assertEqual(pendulum.parse(end_date), pendulum.today("UTC") - timedelta(days=1))
                self.assertTrue(first_release)

                # use release info for other tasks
                release = OrcidRelease(
                    telescope.dag_id,
                    pendulum.parse(start_date),
                    pendulum.parse(end_date),
                    first_release,
                    max_processes=1,
                )

                # Test transfer task
                mock_transfer.return_value = True, 2
                env.run_task(telescope.transfer.__name__)
                mock_transfer.assert_called_once()
                try:
                    self.assertTupleEqual(mock_transfer.call_args[0], (conn.login, conn.password))
                except AssertionError:
                    raise AssertionError("AWS key id and secret not passed correctly to transfer function")
                self.assertDictEqual(
                    mock_transfer.call_args[1],
                    {
                        "aws_bucket": OrcidTelescope.SUMMARIES_BUCKET,
                        "include_prefixes": [],
                        "gc_project_id": self.project_id,
                        "gc_bucket": orcid_bucket,
                        "description": "Transfer ORCID data from airflow " "telescope",
                        "last_modified_since": None,
                    },
                )
                # Upload files to bucket, to mock transfer
                record1 = self.records["0000-0002-9227-8610.xml"]
                record2 = self.records["0000-0002-9228-8514.xml"]
                upload_files_to_cloud_storage(
                    orcid_bucket, [record1["blob"], record2["blob"]], [record1["path"], record2["path"]]
                )
                self.assert_blob_integrity(orcid_bucket, record1["blob"], record1["path"])
                self.assert_blob_integrity(orcid_bucket, record2["blob"], record2["path"])

                # Test that file was downloaded
                env.run_task(telescope.download_transferred.__name__)
                downloaded_hashes = {
                    "0000-0002-9227-8610.xml": "31d17a63cd04cbd5733cafe7f3561cb7",
                    "0000-0002-9228-8514.xml": "0e3426db67d221c9cc53737478ea968c",
                }
                self.assertEqual(2, len(release.download_files))
                for file in release.download_files:
                    self.assert_file_integrity(file, downloaded_hashes[os.path.basename(file)], "md5")

                # Test that files transformed
                env.run_task(telescope.transform.__name__)
                transform_hashes = {"610.jsonl.gz": "33f64619", "514.jsonl.gz": "f1179546"}
                self.assertEqual(2, len(release.transform_files))
                # Sort lines so that gzip crc is always the same
                for file in release.transform_files:
                    with gzip.open(file, "rb") as f_in:
                        lines = sorted(f_in.readlines())
                    with gzip.open(file, "wb") as f_out:
                        f_out.writelines(lines)
                    self.assert_file_integrity(file, transform_hashes[os.path.basename(file)], "gzip_crc")

                # Test that transformed file uploaded
                env.run_task(telescope.upload_transformed.__name__)
                for file in release.transform_files:
                    self.assert_blob_integrity(env.transform_bucket, blob_name(file), file)

                # Test that load partition task is skipped for the first release
                ti = env.run_task(telescope.bq_load_partition.__name__)
                self.assertEqual(ti.state, "skipped")

                # Test delete old task is in success state, without doing anything
                ti = env.run_task(telescope.bq_delete_old.__name__)
                self.assertEqual(ti.state, "success")

                # Test append new creates table
                env.run_task(telescope.bq_append_new.__name__)
                main_table_id, partition_table_id = release.dag_id, f"{release.dag_id}_partitions"
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 2
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

            # second run
            with env.create_dag_run(dag, self.second_execution_date):
                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__)

                # Test list releases task with files available
                ti = env.run_task(telescope.get_release_info.__name__)
                start_date, end_date, first_release = ti.xcom_pull(
                    key=OrcidTelescope.RELEASE_INFO,
                    task_ids=telescope.get_release_info.__name__,
                    include_prior_dates=False,
                )
                self.assertEqual(release.end_date + timedelta(days=1), pendulum.parse(start_date))
                self.assertEqual(pendulum.today("UTC") - timedelta(days=1), pendulum.parse(end_date))
                self.assertFalse(first_release)

                # Use release info for other tasks
                release = OrcidRelease(
                    telescope.dag_id,
                    pendulum.parse(start_date),
                    pendulum.parse(end_date),
                    first_release,
                    max_processes=1,
                )

                # Test transfer task
                mock_transfer.return_value = True, 2
                mock_transfer.reset_mock()
                env.run_task(telescope.transfer.__name__)
                mock_transfer.assert_called_once()
                try:
                    self.assertTupleEqual(mock_transfer.call_args[0], (conn.login, conn.password))
                except AssertionError:
                    raise AssertionError("AWS key id and secret not passed correctly to transfer function")
                self.assertDictEqual(
                    mock_transfer.call_args[1],
                    {
                        "aws_bucket": OrcidTelescope.SUMMARIES_BUCKET,
                        "include_prefixes": [],
                        "gc_project_id": self.project_id,
                        "gc_bucket": orcid_bucket,
                        "description": "Transfer ORCID data from airflow " "telescope",
                        "last_modified_since": release.start_date,
                    },
                )
                # Upload files to bucket, to mock transfer
                record3 = self.records["0000-0002-9229-8514.xml"]
                upload_file_to_cloud_storage(orcid_bucket, record3["blob"], record3["path"])
                self.assert_blob_integrity(orcid_bucket, record3["blob"], record3["path"])

                # Mock response of get_object on last_modified file, mocking lambda file
                with open(self.last_modified_path, "rb") as f:
                    file_bytes = f.read()
                    mock_client().get_object.return_value = {
                        "Body": StreamingBody(io.BytesIO(file_bytes), len(file_bytes))
                    }
                # Test that file was downloaded
                env.run_task(telescope.download_transferred.__name__)
                self.assertEqual(2, len(release.download_files))
                for file in release.download_files:
                    downloaded_hashes = {
                        "0000-0002-9228-8514.xml": "0e3426db67d221c9cc53737478ea968c",
                        "0000-0002-9229-8514.xml": "38472bea0cc72cbefa54f0bf5f98d95f",
                    }
                    self.assert_file_integrity(file, downloaded_hashes[os.path.basename(file)], "md5")

                # Test that files transformed
                env.run_task(telescope.transform.__name__)
                self.assertEqual(1, len(release.transform_files))
                transform_path = release.transform_files[0]
                with gzip.open(transform_path, "rb") as f_in:
                    lines = sorted(f_in.readlines())
                with gzip.open(transform_path, "wb") as f_out:
                    f_out.writelines(lines)
                expected_file_hash = "aab89332"
                self.assert_file_integrity(transform_path, expected_file_hash, "gzip_crc")

                # Test that transformed file uploaded
                env.run_task(telescope.upload_transformed.__name__)
                self.assert_blob_integrity(env.transform_bucket, blob_name(transform_path), transform_path)

                # Test that load partition task creates partition
                env.run_task(telescope.bq_load_partition.__name__)
                main_table_id, partition_table_id = release.dag_id, f"{release.dag_id}_partitions"
                table_id = f"{self.project_id}.{telescope.dataset_id}.{partition_table_id}"
                expected_rows = 2
                self.assert_table_integrity(table_id, expected_rows)

                # Test task deleted rows from main table
                env.run_task(telescope.bq_delete_old.__name__)
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 1
                self.assert_table_integrity(table_id, expected_rows)

                # Test append new adds rows to table
                env.run_task(telescope.bq_append_new.__name__)
                table_id = f"{self.project_id}.{telescope.dataset_id}.{main_table_id}"
                expected_rows = 3
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)

    @patch("academic_observatory_workflows.workflows.orcid_telescope.aws_to_google_cloud_storage_transfer")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.get_aws_conn_info")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.Variable.get")
    def test_transfer(self, mock_variable_get, mock_aws_info, mock_transfer):
        """Test transfer method of the ORCID release.

        :param mock_variable_get: Mock Airflow Variable get() method
        :param mock_aws_info: Mock getting AWS info
        :param mock_transfer: Mock the transfer function called inside release.transfer()
        :return: None.
        """
        mock_variable_get.side_effect = lambda x: {"orcid_bucket": "bucket", "project_id": "project_id"}[x]
        mock_aws_info.return_value = "key_id", "secret_key"
        max_retries = 3

        mock_transfer.return_value = True, 3
        # Test transfer in case of first release
        self.release.first_release = False
        self.release.transfer(max_retries)
        mock_transfer.assert_called_once_with(
            "key_id",
            "secret_key",
            aws_bucket=OrcidTelescope.SUMMARIES_BUCKET,
            include_prefixes=[],
            gc_project_id="project_id",
            gc_bucket="bucket",
            description="Transfer ORCID data from airflow telescope",
            last_modified_since=self.release.start_date,
        )
        mock_transfer.reset_mock()

        # Test transfer in case of later release
        self.release.first_release = True
        self.release.transfer(max_retries)
        mock_transfer.assert_called_once_with(
            "key_id",
            "secret_key",
            aws_bucket=OrcidTelescope.SUMMARIES_BUCKET,
            include_prefixes=[],
            gc_project_id="project_id",
            gc_bucket="bucket",
            description="Transfer ORCID data from airflow telescope",
            last_modified_since=None,
        )
        # Test failed transfer
        mock_transfer.return_value = False, 4
        with self.assertRaises(AirflowException):
            self.release.transfer(max_retries)

        # Test succesful transfer, but no objects were transferred
        mock_transfer.return_value = True, 0
        with self.assertRaises(AirflowSkipException):
            self.release.transfer(max_retries)

    @patch("academic_observatory_workflows.workflows.orcid_telescope.subprocess.Popen")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.get_aws_conn_info")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.write_modified_record_blobs")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.Variable.get")
    def test_download_transferred(self, mock_variable_get, mock_write_blobs, mock_aws_info, mock_subprocess):
        """Test the download_transferred method of the ORCID release.

        :param mock_variable_get: Mock Airflow Variable get() method
        :param mock_write_blobs: Mock the function that writes modified record blobs
        :param mock_aws_info: Mock getting AWS info
        :param mock_subprocess: Mock the subprocess returncode and communicate method
        :return: None.
        """
        mock_variable_get.return_value = "orcid_bucket"
        mock_aws_info.return_value = "key_id", "secret_key"

        mock_subprocess.return_value.returncode = 0
        mock_subprocess.return_value.communicate.return_value = "stdout".encode(), "stderr".encode()

        # Test download in case of first release
        self.release.first_release = True
        self.release.download_transferred()
        mock_write_blobs.assert_not_called()
        self.assertEqual(2, mock_subprocess.call_count)
        mock_subprocess.assert_any_call(
            [
                "gcloud",
                "auth",
                "activate-service-account",
                f'--key-file={os.environ["GOOGLE_APPLICATION_CREDENTIALS"]}',
            ],
            stdout=-1,
            stderr=-1,
        )
        mock_subprocess.assert_called_with(
            [
                "gsutil",
                "-m",
                "-q",
                "cp",
                "-L",
                os.path.join(self.release.download_folder, "cp.log"),
                "-r",
                "gs://orcid_bucket",
                self.release.download_folder,
            ],
            stdout=-1,
            stderr=-1,
        )

        # Test download in case of second release, using modified records file
        self.release.first_release = False
        mock_subprocess.reset_mock()
        with CliRunner().isolated_filesystem():
            with open(self.release.modified_records_path, "w") as f:
                f.write("unit test")
            self.release.download_transferred()
            mock_write_blobs.assert_called_once_with(
                self.release.start_date,
                self.release.end_date,
                "key_id",
                "secret_key",
                "orcid_bucket",
                self.release.modified_records_path,
            )
            self.assertEqual(2, mock_subprocess.call_count)
            mock_subprocess.assert_any_call(
                [
                    "gcloud",
                    "auth",
                    "activate-service-account",
                    f'--key-file={os.environ["GOOGLE_APPLICATION_CREDENTIALS"]}',
                ],
                stdout=-1,
                stderr=-1,
            )
            mock_subprocess.assert_called_with(
                [
                    "gsutil",
                    "-m",
                    "-q",
                    "cp",
                    "-L",
                    os.path.join(self.release.download_folder, "cp.log"),
                    "-I",
                    self.release.download_folder,
                ],
                stdin=ANY,
                stdout=-1,
                stderr=-1,
            )
            self.assertEqual(self.release.modified_records_path, mock_subprocess.call_args[1]["stdin"].name)

            # Test download when first subprocess fails
            mock_subprocess.return_value.returncode = -1
            mock_subprocess.return_value.communicate.return_value = "stdout".encode(), "stderr".encode()
            with self.assertRaises(AirflowException):
                self.release.download_transferred()

            # Test download when second subprocess fails
            def communicate():
                return "stdout".encode(), "stderr".encode()

            mock_subprocess.side_effect = [
                SimpleNamespace(communicate=communicate, returncode=0),
                SimpleNamespace(communicate=communicate, returncode=-1),
            ]
            with self.assertRaises(AirflowException):
                self.release.download_transferred()

    @patch("academic_observatory_workflows.workflows.orcid_telescope.Variable.get")
    def test_transform_single_file(self, mock_variable_get):
        """Test the transform_single_file method.

        :return: None.
        """
        with CliRunner().isolated_filesystem() as t:
            mock_variable_get.return_value = os.path.join(t, "data")

            file_name = "0000-0002-9228-8514.xml"
            transform_folder = self.release.transform_folder
            file_dir = os.path.join(self.release.transform_folder, file_name[-7:-4])
            transform_path = os.path.join(file_dir, os.path.splitext(file_name)[0] + ".jsonl")

            # Test standard record
            with open(file_name, "w") as f:
                f.write(
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    '<record:record path="/0000-0002-9227-8514">'
                    "<common:orcid-identifier>"
                    "<common:path>0000-0002-9227-8514</common:path>"
                    "</common:orcid-identifier>"
                    "</record:record>"
                )
            transform_single_file(file_name, transform_folder)
            self.assert_file_integrity(transform_path, "6d7dbc0fc69db96025b82c018b3d6305", "md5")

            # Test transform standard record is skipped, because file already exists
            transform_single_file(file_name, transform_folder)
            self.assert_file_integrity(transform_path, "6d7dbc0fc69db96025b82c018b3d6305", "md5")
            os.remove(transform_path)

            # Test record with error
            with open(file_name, "w") as f:
                f.write(
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    '<error:error path="/0000-0002-9227-8514">'
                    "<common:orcid-identifier>"
                    "<common:path>0000-0002-9227-8514</common:path>"
                    "</common:orcid-identifier>"
                    "</error:error>"
                )
            transform_single_file(file_name, transform_folder)
            self.assert_file_integrity(transform_path, "6d7dbc0fc69db96025b82c018b3d6305", "md5")
            os.remove(transform_path)

            # Test invalid record
            with open(file_name, "w") as f:
                f.write(
                    '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
                    '<invalid:invalid> test="test">'
                    "</invalid:invalid>"
                )
            with self.assertRaises(AirflowException):
                transform_single_file(file_name, transform_folder)

    @patch("academic_observatory_workflows.workflows.orcid_telescope.Variable.get")
    @patch.object(BaseHook, "get_connection")
    @patch("academic_observatory_workflows.workflows.orcid_telescope.storage_bucket_exists")
    def test_check_dependencies(self, mock_bucket_exists, mock_conn_get, mock_variable_get):
        """Test the check_dependencies task

        :param mock_bucket_exists: Mock output of storage_bucket_exists function
        :param mock_conn_get: Mock Airflow get_connection method
        :param mock_variable_get: Mock Airflow Variable get() method
        :return:
        """
        mock_variable_get.return_value = "orcid_bucket"
        mock_conn_get.return_value = "orcid"

        # Test that all dependencies are specified: no error should be thrown
        mock_bucket_exists.return_value = True
        OrcidTelescope().check_dependencies()

        # Test that dependency is missing, no existing storage bucket
        mock_bucket_exists.return_value = False
        with self.assertRaises(AirflowException):
            OrcidTelescope().check_dependencies()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs, Tuan Chien

import logging
import os
import unittest
from pathlib import Path
from unittest.mock import MagicMock, PropertyMock, patch

import pendulum
import vcr
from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.grid_telescope import (
    GridRelease,
    GridTelescope,
    list_grid_records,
)
from airflow.exceptions import AirflowException
from click.testing import CliRunner
from observatory.platform.utils.file_utils import _hash_file, gzip_file_crc
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.workflow_utils import blob_name, table_ids_from_path


class MockResponse:
    def __init__(self):
        self.text = '[{"published_date": "20210101", "id":12345, "title":"no date in here"}]'


class MockSession:
    def get(self, *args, **kwargs):
        return MockResponse()


class MockTaskInstance:
    def __init__(self, records):
        """Construct a MockTaskInstance. This mocks the airflow TaskInstance and is passed as a keyword arg to the
        make_release function.
        :param records: List of record info, returned as value during xcom_pull
        """
        self.records = records

    def xcom_pull(self, key: str, task_ids: str, include_prior_dates: bool):
        """Mock xcom_pull method of airflow TaskInstance.
        :param key: -
        :param task_ids: -
        :param include_prior_dates: -
        :return: Records list
        """
        return self.records


def side_effect(arg):
    values = {
        "project_id": "project",
        "download_bucket_name": "download-bucket",
        "transform_bucket_name": "transform-bucket",
        "data_path": "data",
        "data_location": "US",
    }
    return values[arg]


@patch("observatory.platform.utils.workflow_utils.Variable.get")
class TestGridTelescope(unittest.TestCase):
    """Tests for the functions used by the GRID telescope"""

    def __init__(
        self,
        *args,
        **kwargs,
    ):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestGridTelescope, self).__init__(*args, **kwargs)
        # Telescope instance
        self.grid = GridTelescope()

        # Paths
        self.vcr_cassettes_path = test_fixtures_folder("grid")
        self.list_grid_records_path = os.path.join(self.vcr_cassettes_path, "list_grid_releases.yaml")

        # Contains GRID releases 2015-09-22 and 2015-10-09 (format for both is .csv and .json files)
        with patch("observatory.platform.utils.workflow_utils.Variable.get") as mock_variable_get:
            mock_variable_get.side_effect = side_effect
            self.grid_run_2015_10_18 = {
                "start_date": pendulum.datetime(2015, 10, 11),
                "end_date": pendulum.datetime(2015, 10, 18),
                "records": [
                    {"article_ids": [1570968, 1570967], "release_date": "2015-10-09"},
                    {"article_ids": [1553266, 1553267], "release_date": "2015-09-22"},
                ],
                # there are 2 releases in this run, but use only 1 for testing
                "release": GridRelease(
                    self.grid.dag_id, ["1553266", "1553267"], pendulum.parse("2015-09-22T00:00:00+00:00")
                ),
                "download_path": os.path.join(self.vcr_cassettes_path, "grid_2015-09-22.yaml"),
                "download_hash": "c6fd33fd31b6699a2f19622f0283f4f1",
                "extract_hash": "c6fd33fd31b6699a2f19622f0283f4f1",
                "transform_crc": "eb66ae78",
            }

            # Contains GRID release 2020-03-15 (format is a .zip file, which is more common)
            self.grid_run_2020_03_27 = {
                "start_date": pendulum.datetime(2020, 3, 20),
                "end_date": pendulum.datetime(2020, 3, 27),
                "records": [{"article_ids": [12022722], "release_date": "2020-03-15T00:00:00+00:00"}],
                "release": GridRelease(self.grid.dag_id, ["12022722"], pendulum.parse("2020-03-15T00:00:00+00:00")),
                "download_path": os.path.join(self.vcr_cassettes_path, "grid_2020-03-15.yaml"),
                "download_hash": "3d300affce1666ac50b8d945c6ca4c5a",
                "extract_hash": "5aff68e9bf72e846a867e91c1fa206a0",
                "transform_crc": "77bc8585",
            }

        self.grid_runs = [self.grid_run_2015_10_18, self.grid_run_2020_03_27]

        # Turn logging to warning because vcr prints too much at info level
        logging.basicConfig()
        logging.getLogger().setLevel(logging.WARNING)

    def test_ctor(self, mock_variable_get):
        """Cover case where airflow_vars is given."""
        telescope = GridTelescope(airflow_vars=[])
        self.assertEqual(telescope.airflow_vars, list(["transform_bucket"]))

    def test_list_grid_records(self, mock_variable_get):
        """Check that list grid records returns a list of dictionaries with records in the correct format.

        :param mock_variable_get: Mock result of airflow's Variable.get() function
        :return: None.
        """
        with vcr.use_cassette(self.list_grid_records_path):
            start_date = self.grid_run_2015_10_18["start_date"]
            end_date = self.grid_run_2015_10_18["end_date"]
            records = list_grid_records(start_date, end_date, GridTelescope.GRID_DATASET_URL)
            self.assertEqual(self.grid_run_2015_10_18["records"], records)

    def test_list_grid_records_bad_title(self, mock_variable_get):
        """Check exception raised when invalid title given."""

        with patch(
            "academic_observatory_workflows.workflows.grid_telescope.retry_session", return_value=MockSession()
        ) as _:
            start_date = pendulum.datetime(2020, 1, 1)
            end_date = pendulum.datetime(2022, 1, 1)
            self.assertRaises(ValueError, list_grid_records, start_date, end_date, "")

    def test_list_releases(self, mock_variable_get):
        """Test list_releases."""

        ti = MagicMock()
        with patch("academic_observatory_workflows.workflows.grid_telescope.list_grid_records") as m_list_grid_records:
            m_list_grid_records.return_value = []
            telescope = GridTelescope()
            result = telescope.list_releases(execution_date=pendulum.now(), next_execution_date=pendulum.now())
            self.assertEqual(result, False)

            m_list_grid_records.return_value = [1]
            telescope = GridTelescope()
            result = telescope.list_releases(execution_date=pendulum.now(), next_execution_date=pendulum.now(), ti=ti)
            self.assertEqual(result, True)

    def test_make_release(self, mock_variable_get):
        """Check that make_release returns a list of GridRelease instances.

        :param mock_variable_get: Mock result of airflow's Variable.get() function
        :return: None.
        """
        mock_variable_get.side_effect = side_effect

        for run in self.grid_runs:
            records = run["records"]

            releases = self.grid.make_release(ti=MockTaskInstance(records))
            self.assertIsInstance(releases, list)
            for release in releases:
                self.assertIsInstance(release, GridRelease)

    def test_download_release(self, mock_variable_get):
        """Download two specific GRID releases and check they have the expected md5 sum.

        :param mock_variable_get: Mock result of airflow's Variable.get() function
        :return:
        """
        mock_variable_get.side_effect = side_effect

        with CliRunner().isolated_filesystem():
            for run in self.grid_runs:
                release = run["release"]
                with vcr.use_cassette(run["download_path"]):
                    downloads = release.download()
                    # Check that returned downloads has correct length
                    self.assertEqual(1, len(downloads))

                    # Check that file has expected hash
                    file_path = downloads[0]
                    self.assertTrue(os.path.exists(file_path))
                    self.assertEqual(run["download_hash"], _hash_file(file_path, algorithm="md5"))

    def test_extract_release(self, mock_variable_get):
        """Test that the GRID releases are extracted as expected, both for an unzipped json file and a zip file.

        :param mock_variable_get: Mock result of airflow's Variable.get() function
        :return: None.
        """
        mock_variable_get.side_effect = side_effect

        with CliRunner().isolated_filesystem():
            for run in self.grid_runs:
                release = run["release"]
                with vcr.use_cassette(run["download_path"]):
                    release.download()
                    release.extract()

                    self.assertEqual(1, len(release.extract_files))
                    self.assertEqual(run["extract_hash"], _hash_file(release.extract_files[0], algorithm="md5"))

    def test_transform_release(self, mock_variable_get):
        """Test that the GRID releases are transformed as expected.

        :param mock_variable_get: Mock result of airflow's Variable.get() function
        :return: None.
        """
        mock_variable_get.side_effect = side_effect

        with CliRunner().isolated_filesystem():
            for run in self.grid_runs:
                release = run["release"]
                with vcr.use_cassette(run["download_path"]):
                    release.download()
                    release.extract()
                    release.transform()

                    self.assertEqual(1, len(release.transform_files))
                    self.assertEqual(run["transform_crc"], gzip_file_crc(release.transform_files[0]))


class TestGridRelease(unittest.TestCase):
    @patch(
        "academic_observatory_workflows.workflows.grid_telescope.GridRelease.extract_folder", new_callable=PropertyMock
    )
    @patch(
        "academic_observatory_workflows.workflows.grid_telescope.GridRelease.download_files", new_callable=PropertyMock
    )
    def test_extract_not_zip_file(self, m_download_files, m_extract_folder):
        with CliRunner().isolated_filesystem():
            Path("file.zip").touch()

            release = GridRelease(dag_id="dag", article_ids=[], release_date=pendulum.now())
            m_download_files.return_value = ["file.zip"]
            m_extract_folder.return_value = "."
            release.extract()

    @patch(
        "academic_observatory_workflows.workflows.grid_telescope.GridRelease.extract_files", new_callable=PropertyMock
    )
    def test_transform_multiple_extract(self, m_extract_files):
        m_extract_files.return_value = ["1", "2"]

        with CliRunner().isolated_filesystem():
            release = GridRelease(dag_id="dag", article_ids=[], release_date=pendulum.now())
            self.assertRaises(AirflowException, release.transform)


class TestGridTelescopeDag(ObservatoryTestCase):
    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """
        super().__init__(*args, **kwargs)

        self.project_id = os.environ["TEST_GCP_PROJECT_ID"]
        self.data_location = os.environ["TEST_GCP_DATA_LOCATION"]

        # Paths
        self.vcr_cassettes_path = os.path.join(test_fixtures_folder(), "grid")
        self.list_grid_records_path = os.path.join(self.vcr_cassettes_path, "list_grid_releases.yaml")
        self.grid_path = os.path.join(self.vcr_cassettes_path, "grid_2015-09-22.yaml")

    def setup_observatory_environment(self):
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        self.dataset_id = env.add_dataset()
        return env

    def test_dag_structure(self):
        """Test that the GRID DAG has the correct structure.

        :return: None
        """
        telescope = GridTelescope()
        dag = telescope.make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["list_releases"],
                "list_releases": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["extract"],
                "extract": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the GRID DAG can be loaded from a DAG bag.

        :return: None
        """

        env = ObservatoryEnvironment(self.project_id, self.data_location)
        with env.create():
            dag_file = os.path.join(module_file_path("academic_observatory_workflows.dags"), "grid_telescope.py")
            self.assert_dag_load("grid", dag_file)

    def test_telescope(self):
        """Test running the telescope. Functional test."""

        env = self.setup_observatory_environment()
        telescope = GridTelescope(dag_id="grid", dataset_id=self.dataset_id)
        dag = telescope.make_dag()
        execution_date = pendulum.datetime(year=2015, month=9, day=22)

        # Create the Observatory environment and run tests
        with env.create():
            with env.create_dag_run(dag, execution_date):
                # Check dependencies
                env.run_task(telescope.check_dependencies.__name__, dag, execution_date)

                # List releases
                with patch(
                    "academic_observatory_workflows.workflows.grid_telescope.list_grid_records"
                ) as m_list_grid_records:
                    m_list_grid_records.return_value = [
                        {"article_ids": [1553266, 1553267], "release_date": "2015-10-09"},
                    ]
                    ti = env.run_task(telescope.list_releases.__name__, dag, execution_date)

                # Test list releases
                available_releases = ti.xcom_pull(
                    key=GridTelescope.RELEASE_INFO,
                    task_ids=telescope.list_releases.__name__,
                    include_prior_dates=False,
                )
                self.assertEqual(len(available_releases), 1)

                # Download
                with vcr.use_cassette(self.grid_path):
                    env.run_task(telescope.download.__name__, dag, execution_date)

                # Test download
                release = GridRelease(
                    dag_id="grid",
                    article_ids=[1553266, 1553267],
                    release_date=pendulum.datetime(2015, 10, 9),
                )
                self.assertEqual(len(release.download_files), 1)

                # upload_downloaded
                env.run_task(telescope.upload_downloaded.__name__, dag, execution_date)

                # Test upload_downloaded
                for file in release.download_files:
                    self.assert_blob_integrity(env.download_bucket, blob_name(file), file)

                # extract
                env.run_task(telescope.extract.__name__, dag, execution_date)

                # Test extract
                self.assertEqual(len(release.extract_files), 1)

                # transform
                env.run_task(telescope.transform.__name__, dag, execution_date)

                # Test transform
                self.assertEqual(len(release.transform_files), 1)

                # upload_transformed
                env.run_task(telescope.upload_transformed.__name__, dag, execution_date)

                # Test upload_transformed
                for file in release.transform_files:
                    self.assert_blob_integrity(env.transform_bucket, blob_name(file), file)

                # bq_load
                env.run_task(telescope.bq_load.__name__, dag, execution_date)

                # Test bq_load
                # Will only check table exists rather than validate data.
                for file in release.transform_files:
                    table_id, _ = table_ids_from_path(file)
                    suffix = release.release_date.format("YYYYMMDD")
                    table_id = f"{self.project_id}.{self.dataset_id}.{table_id}{suffix}"
                    expected_rows = 48987
                    self.assert_table_integrity(table_id, expected_rows)

                # cleanup
                env.run_task(telescope.cleanup.__name__, dag, execution_date)

                # Test cleanup
                # Test that all telescope data deleted
                download_folder, extract_folder, transform_folder = (
                    release.download_folder,
                    release.extract_folder,
                    release.transform_folder,
                )
                env.run_task(telescope.cleanup.__name__, dag, execution_date)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Tuan Chien


import json
import logging
import unittest
from unittest.mock import patch

import pendulum
from airflow.models import Connection
from click.testing import CliRunner

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.web_of_science_telescope import WosJsonParser, WosUtility


# class TestRecordWosVcr(unittest.TestCase):
#     """ A test that records the WoS response into a vcr cassette. This is in lieu of creating a separate program to
#     do the recording. Comment/uncomment as needed.
#     """
#
#     def record(self, client, inst_id, period):
#         query = WosUtility.build_query(inst_id, period)
#         WosUtility.make_query(client, query)
#
#     def test_record(self):
#         period = pendulum.Period(pendulum.date(1965, 10, 1), pendulum.date(1965, 10, 1))
#
#         with vcr.use_cassette(f'/tmp/wos_cassette.yaml'):
#             with WosClient('login', 'password') as client:
#                 self.record(client, ['Curtin University'], period)


class TestWosParse(unittest.TestCase):
    """Test the various wos parsing utilities."""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestWosParse, self).__init__(*args, **kwargs)

        # Paths
        self.work_dir = "../../../tests/observatory/dags/workflows"
        self.wos_2020_10_01_json_path = test_fixtures_folder("web_of_science", "wos-2020-10-01.json")
        with open(self.wos_2020_10_01_json_path, "r") as f:
            self.data = json.load(f)
        self.harvest_datetime = pendulum.now().isoformat()
        self.release_date = pendulum.date(2020, 10, 1).isoformat()

    def test_get_identifiers(self):
        """Extract identifiers"""
        data = self.data[0]
        identifiers = WosJsonParser.get_identifiers(data)
        self.assertEqual(len(identifiers), 10)
        self.assertEqual(identifiers["uid"], "WOS:000000000000000")
        self.assertEqual(identifiers["issn"], "0000-0000")
        self.assertEqual(identifiers["eissn"], "0000-0000")
        self.assertEqual(identifiers["doi"], "10.0000/j.gaz.2020.01.001")

    def test_get_pub_info(self):
        """Extract publication info"""
        data = self.data[0]
        pub_info = WosJsonParser.get_pub_info(data)
        self.assertEqual(pub_info["sort_date"], "2020-01-01")
        self.assertEqual(pub_info["pub_type"], "Journal")
        self.assertEqual(pub_info["page_count"], 2)
        self.assertEqual(pub_info["source"], "JUPITER GAZETTE")
        self.assertEqual(pub_info["doc_type"], "Article")
        self.assertEqual(pub_info["publisher"], "JUPITER PUBLISHING LTD")
        self.assertEqual(pub_info["publisher_city"], "SPRINGFIELD")

    def test_get_title(self):
        """Extract title"""
        data = self.data[0]
        title = WosJsonParser.get_title(data)
        truth = (
            "The habitats of endangered hypnotoads on the southern oceans of Europa: a Ophiophagus hannah perspective"
        )
        self.assertEqual(title, truth)

    def test_get_names(self):
        """Extract name information, e.g. authors"""
        data = self.data[0]
        names = WosJsonParser.get_names(data)
        self.assertEqual(len(names), 3)

        entry = names[0]
        self.assertEqual(entry["seq_no"], 1)
        self.assertEqual(entry["role"], "author")
        self.assertEqual(entry["first_name"], "Big Eaty")
        self.assertEqual(entry["last_name"], "Snake")
        self.assertEqual(entry["wos_standard"], "Snake, BE")
        self.assertEqual(entry["daisng_id"], "101010")
        self.assertEqual(entry["full_name"], "Snake, Big Eaty")
        self.assertEqual(entry["orcid"], "0000-0000-0000-0001")
        self.assertEqual(entry["r_id"], "D-0000-2000")

        entry = names[1]
        self.assertEqual(entry["seq_no"], 2)
        self.assertEqual(entry["role"], "author")
        self.assertEqual(entry["first_name"], "Hypno")
        self.assertEqual(entry["last_name"], "Toad")
        self.assertEqual(entry["wos_standard"], "Toad, H")
        self.assertEqual(entry["daisng_id"], "100000")
        self.assertEqual(entry["full_name"], "Toad, Hypno")
        self.assertEqual(entry["orcid"], "0000-0000-0000-0002")
        self.assertEqual(entry["r_id"], "H-0000-2001")

        entry = names[2]
        self.assertEqual(entry["seq_no"], 3)
        self.assertEqual(entry["role"], "author")
        self.assertEqual(entry["first_name"], "Great")
        self.assertEqual(entry["last_name"], "Historian")
        self.assertEqual(entry["wos_standard"], "Historian, G")
        self.assertEqual(entry["daisng_id"], "200000")
        self.assertEqual(entry["full_name"], "Historian, Great")
        self.assertEqual(entry["orcid"], "0000-0000-0000-0003")
        self.assertEqual(entry["r_id"], None)

    def test_get_languages(self):
        """Extract language information"""
        data = self.data[0]
        languages = WosJsonParser.get_languages(data)
        self.assertEqual(len(languages), 1)
        self.assertEqual(languages[0]["type"], "primary")
        self.assertEqual(languages[0]["name"], "Mindwaves")

    def test_get_refcount(self):
        """Extract reference count"""
        data = self.data[0]
        refs = WosJsonParser.get_refcount(data)
        self.assertEqual(refs, 10000)

    def test_get_abstract(self):
        """Extract the abstract"""
        data = self.data[0]
        abstract = WosJsonParser.get_abstract(data)
        self.assertEqual(len(abstract), 1)
        head = abstract[0][0:38]
        truth = "Jupiter hypnotoads lead mysterious liv"
        self.assertEqual(head, truth)
        self.assertEqual(len(abstract[0]), 169)

    def test_get_keyword(self):
        """Extract keywords and keywords plus if available"""
        data = self.data[0]
        keywords = WosJsonParser.get_keyword(data)
        self.assertEqual(len(keywords), 15)
        word_list = [
            "Jupiter",
            "Toads",
            "Snakes",
            "JPT",
            "JPS",
            "WORD1",
            "WORD2",
            "WORD3",
            "WORD4",
            "WORD5",
            "WORD6",
            "WORD7",
            "WORD8",
            "WORD9",
            "WORD0",
        ]
        self.assertListEqual(keywords, word_list)

    def test_get_conference(self):
        """Extract conference name"""
        data = self.data[0]
        conf = WosJsonParser.get_conference(data)
        name = "Annual Jupiter Meeting of the Minds"
        self.assertEqual(len(conf), 1)
        self.assertEqual(conf[0]["name"], name)
        self.assertEqual(conf[0]["id"], 12345)

    def test_get_fund_ack(self):
        """Extract funding information"""
        data = self.data[0]
        fund_ack = WosJsonParser.get_fund_ack(data)
        truth = "The authors would like to thank all life in the universe for not making us extinct yet."
        self.assertEqual(len(fund_ack["text"]), 1)
        self.assertEqual(fund_ack["text"][0], truth)
        self.assertEqual(len(fund_ack["grants"]), 1)
        self.assertEqual(fund_ack["grants"][0]["agency"], "Jupiter research council")
        self.assertEqual(len(fund_ack["grants"][0]["ids"]), 1)
        self.assertEqual(fund_ack["grants"][0]["ids"][0], "JP00000000HT1")

    def test_get_categories(self):
        """Extract WoS categories"""
        data = self.data[0]
        categories = WosJsonParser.get_categories(data)
        self.assertEqual(len(categories["headings"]), 1)
        self.assertEqual(len(categories["subheadings"]), 1)
        self.assertEqual(len(categories["subjects"]), 3)

        self.assertEqual(categories["headings"][0], "Hynology")
        self.assertEqual(categories["subheadings"][0], "Zoology")

        self.assertDictEqual(
            categories["subjects"][0], {"ascatype": "traditional", "code": "XX", "text": "Jupiter Toads"}
        )
        self.assertDictEqual(
            categories["subjects"][1], {"ascatype": "traditional", "code": "X", "text": "Jupiter life"}
        )
        self.assertDictEqual(
            categories["subjects"][2], {"ascatype": "extended", "code": None, "text": "Jupiter Science"}
        )

    def test_get_orgs(self):
        """Extract Wos organisations"""
        data = self.data[0]
        orgs = WosJsonParser.get_orgs(data)
        self.assertEqual(len(orgs), 1)
        self.assertEqual(orgs[0]["city"], "Springfield")
        self.assertEqual(orgs[0]["state"], "SF")
        self.assertEqual(orgs[0]["country"], "Jupiter")
        self.assertEqual(orgs[0]["org_name"], "Generic University")
        self.assertEqual(len(orgs[0]["suborgs"]), 2)
        self.assertEqual(orgs[0]["suborgs"][0], "Centre of Excellence for Extraterrestrial Telepathic Studies")
        self.assertEqual(orgs[0]["suborgs"][1], "Zoology")
        self.assertEqual(len(orgs[0]["names"]), 3)
        self.assertEqual(orgs[0]["names"][0]["first_name"], "Big Eaty")
        self.assertEqual(orgs[0]["names"][0]["last_name"], "Snake")
        self.assertEqual(orgs[0]["names"][0]["daisng_id"], "101010")
        self.assertEqual(orgs[0]["names"][0]["full_name"], "Snake, Big Eaty")
        self.assertEqual(orgs[0]["names"][0]["wos_standard"], "Snake, BE")
        self.assertEqual(orgs[0]["names"][1]["first_name"], "Hypno")
        self.assertEqual(orgs[0]["names"][1]["last_name"], "Toad")
        self.assertEqual(orgs[0]["names"][1]["daisng_id"], "100000")
        self.assertEqual(orgs[0]["names"][1]["full_name"], "Toad, Hypno")
        self.assertEqual(orgs[0]["names"][1]["wos_standard"], "Toad, H")
        self.assertEqual(orgs[0]["names"][2]["first_name"], "Great")
        self.assertEqual(orgs[0]["names"][2]["last_name"], "Historian")
        self.assertEqual(orgs[0]["names"][2]["daisng_id"], "200000")
        self.assertEqual(orgs[0]["names"][2]["full_name"], "Historian, Great")
        self.assertEqual(orgs[0]["names"][2]["wos_standard"], "Historian, G")

    def test_parse_json(self):
        """Test whether the json file can be parsed into fields correctly."""

        self.assertEqual(len(self.data), 1)
        entry = self.data[0]

        wos_inst_id = ["Generic University"]
        entry = WosJsonParser.parse_json(entry, self.harvest_datetime, self.release_date, wos_inst_id)
        self.assertEqual(entry["harvest_datetime"], self.harvest_datetime)
        self.assertEqual(entry["release_date"], self.release_date)
        self.assertEqual(entry["identifiers"]["uid"], "WOS:000000000000000")
        self.assertEqual(entry["pub_info"]["pub_type"], "Journal")
        self.assertEqual(
            entry["title"],
            "The habitats of endangered hypnotoads on the southern oceans of Europa: a Ophiophagus hannah perspective",
        )
        self.assertEqual(entry["names"][0]["first_name"], "Big Eaty")
        self.assertEqual(entry["languages"][0]["name"], "Mindwaves")
        self.assertEqual(entry["ref_count"], 10000)
        self.assertEqual(len(entry["abstract"][0]), 169)
        self.assertEqual(len(entry["keywords"]), 15)
        self.assertEqual(len(entry["conferences"]), 1)
        self.assertEqual(entry["fund_ack"]["grants"][0]["ids"][0], "JP00000000HT1")
        self.assertEqual(entry["categories"]["headings"][0], "Hynology")
        self.assertEqual(len(entry["orgs"]), 1)


class TestWos(unittest.TestCase):
    """Test the WosTelescope."""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestWos, self).__init__(*args, **kwargs)

        logging.basicConfig()
        logging.getLogger().setLevel(logging.WARNING)

        # Connection details
        self.conn = Connection()
        self.conn.conn_id = "wos_curtin"
        self.conn.login = "test"
        self.conn.password = "test"
        self.conn.extra = '{\r\n "start_date": "2019-07-01",\r\n "id": "Curtin University" \r\n}'

        # Paths
        self.work_dir = "../../../tests/observatory/dags/workflows"

    @patch("academic_observatory_workflows.workflows.web_of_science_telescope.WosClient")
    @patch("academic_observatory_workflows.workflows.web_of_science_telescope.WosUtility.make_query", return_value=[""])
    def test_download_wos_snapshot(self, mock_query, mock_client):
        """Test whether we can successfully download and save a snapshot."""

        with CliRunner().isolated_filesystem():
            dag_start = pendulum.datetime(2019, 11, 1)
            wos_inst_id = ["Curtin University"]
            files = WosUtility.download_wos_snapshot(".", self.conn, wos_inst_id, dag_start, "sequential")
            self.assertEqual(5, len(files))
            self.assertEqual(mock_query.call_count, 5)
            self.assertEqual(mock_client.call_count, 1)



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: James Diprose, Aniek Roelofs

from __future__ import annotations

import json
import logging
import os
import re
from shutil import copyfile
from typing import List
from zipfile import BadZipFile, ZipFile

import pendulum
from academic_observatory_workflows.config import schema_folder as default_schema_folder
from airflow.exceptions import AirflowException
from airflow.models.taskinstance import TaskInstance
from google.cloud.bigquery import SourceFormat
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.data_utils import get_file
from observatory.platform.utils.file_utils import list_to_jsonl_gz
from observatory.platform.utils.url_utils import retry_session
from observatory.platform.utils.workflow_utils import upload_files_from_list
from observatory.platform.workflows.snapshot_telescope import (
    SnapshotRelease,
    SnapshotTelescope,
)


class GridRelease(SnapshotRelease):
    def __init__(self, dag_id: str, article_ids: List[str], release_date: pendulum.DateTime):
        """Construct a GridRelease.

        :param article_ids: the titles of the Figshare articles.
        :param release_date: the release date.
        """

        self.article_ids = article_ids
        download_files_regex = dag_id + "\.[a-zA-Z]+"
        extract_files_regex = "grid.json"
        transform_files_regex = f"{dag_id}.jsonl.gz"

        super().__init__(dag_id, release_date, download_files_regex, extract_files_regex, transform_files_regex)

    @property
    def transform_path(self) -> str:
        return os.path.join(self.transform_folder, f"{self.dag_id}.jsonl.gz")

    def download(self, timeout: float = 30.0) -> List[str]:
        """Downloads an individual GRID release from Figshare.

        :param timeout: the timeout in seconds when calling the Figshare API.
        :return: the paths on the system of the downloaded files.
        """

        downloads = []
        for article_id in self.article_ids:
            response = retry_session().get(
                GridTelescope.GRID_FILE_URL.format(article_id=article_id),
                timeout=timeout,
                headers={"Accept-encoding": "gzip"},
            )
            article_files = json.loads(response.text)

            for i, article_file in enumerate(article_files):
                real_file_name = article_file["name"]
                supplied_md5 = article_file["supplied_md5"]
                download_url = article_file["download_url"]
                file_type = os.path.splitext(real_file_name)[1]

                if file_type == ".csv":
                    continue

                # Download
                logging.info(f"Downloading file: {real_file_name}, md5: {supplied_md5}, url: {download_url}")
                file_path, updated = get_file(
                    f"{self.dag_id}{file_type}",
                    download_url,
                    md5_hash=supplied_md5,
                    cache_subdir="",
                    cache_dir=self.download_folder,
                )
                downloads.append(file_path)

        return downloads

    def extract(self) -> None:
        """Extract a single GRID release to a given extraction path. The release will be extracted into the following
        directory structure: extraction_path/file_name (without extension).

        If the release is a .zip file, it will be extracted, otherwise it will be copied to a directory within the
        extraction path.
        :return: None.
        """

        logging.info(f"Download files {self.download_files}")
        # Extract files
        for file_path in self.download_files:
            # Extract zip files
            if file_path.endswith(".zip"):
                unzip_folder_path = self.extract_folder
                logging.info(f"Extracting file: {file_path}")
                try:
                    with ZipFile(file_path) as zip_file:
                        zip_file.extractall(unzip_folder_path)
                except BadZipFile:
                    logging.error("Not a zip file")
                logging.info(f"File extracted to: {unzip_folder_path}")
            else:
                # File is already uncompressed (.json or .csv), so make a directory and copy it into it
                output_file_path = os.path.join(self.extract_folder, os.path.basename(file_path))
                copyfile(file_path, output_file_path)
                logging.info(f"File saved to: {output_file_path}")

    def transform(self) -> str:
        """Transform an extracted GRID release .json file into json lines format and gzip the result.

        :return: the GRID version, the file name and the file path.
        """

        extract_files = self.extract_files

        # Only process one JSON file
        if len(extract_files) == 1:
            release_json_file = extract_files[0]
            logging.info(f"Transforming file: {release_json_file}")

        else:
            raise AirflowException(f"{len(extract_files)} extracted grid.json file found: {extract_files}")

        with open(release_json_file) as json_file:
            # Load GRID release JSON file
            data = json.load(json_file)
            version = data["version"]
            institutes = data["institutes"]

            # Transform GRID release into JSON Lines format saving in memory buffer
            # Save in memory buffer to gzipped file
            list_to_jsonl_gz(self.transform_path, institutes)

        return version


def list_grid_records(
    start_date: pendulum.DateTime, end_date: pendulum.DateTime, grid_dataset_url: str, timeout: float = 30.0
) -> List[dict]:
    """List all GRID records available on Figshare between two dates.

    :param timeout: the number of seconds to wait until timing out.
    :return: the list of GRID releases with required variables stored as a dictionary.
    """

    response = retry_session().get(grid_dataset_url, timeout=timeout, headers={"Accept-encoding": "gzip"})
    response_json = json.loads(response.text)

    records: List[dict] = []
    release_articles = {}
    for item in response_json:
        published_date: pendulum.DateTime = pendulum.parse(item["published_date"])

        if start_date <= published_date < end_date:
            article_id = item["id"]
            title = item["title"]

            # Parse date:
            # The publish date is not used as the release date because the dataset is often
            # published after the release date
            date_matches = re.search("([0-9]{4}\-[0-9]{2}\-[0-9]{2})", title)
            if date_matches is None:
                raise ValueError(f"No release date found in GRID title: {title}")
            release_date = date_matches[0]

            try:
                release_articles[release_date].append(article_id)
            except KeyError:
                release_articles[release_date] = [article_id]

    for release_date in release_articles:
        article_ids = release_articles[release_date]
        records.append({"article_ids": article_ids, "release_date": release_date})
    return records


class GridTelescope(SnapshotTelescope):
    """
    The Global Research Identifier Database (GRID): https://grid.ac/

    Saved to the BigQuery table: <project_id>.digital_science.gridYYYYMMDD
    """

    DAG_ID = "grid"
    DATASET_ID = "digital_science"
    GRID_FILE_URL = "https://api.figshare.com/v2/articles/{article_id}/files"
    GRID_DATASET_URL = "https://api.figshare.com/v2/collections/3812929/articles?page_size=1000"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2015, 9, 1),
        schedule_interval: str = "@weekly",
        dataset_id: str = DATASET_ID,
        schema_folder: str = default_schema_folder(),
        source_format: str = SourceFormat.NEWLINE_DELIMITED_JSON,
        dataset_description: str = "Datasets provided by Digital Science: https://www.digital-science.com/",
        catchup: bool = True,
        airflow_vars: List = None,
    ):
        """Construct a GridTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the BigQuery dataset id.
        :param schema_folder: the SQL schema path.
        :param source_format: the format of the data to load into BigQuery.
        :param dataset_description: description for the BigQuery dataset.
        :param catchup: whether to catchup the DAG or not.
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        """

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            schema_folder,
            source_format=source_format,
            dataset_description=dataset_description,
            catchup=catchup,
            airflow_vars=airflow_vars,
        )

        self.add_setup_task_chain([self.check_dependencies, self.list_releases])
        self.add_task_chain(
            [
                self.download,
                self.upload_downloaded,
                self.extract,
                self.transform,
                self.upload_transformed,
                self.bq_load,
                self.cleanup,
            ]
        )

    def make_release(self, **kwargs) -> List[GridRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are
        passed to this argument.
        :return: A list of grid release instances
        """

        ti: TaskInstance = kwargs["ti"]
        records = ti.xcom_pull(
            key=GridTelescope.RELEASE_INFO, task_ids=self.list_releases.__name__, include_prior_dates=False
        )
        releases = []
        for record in records:
            article_ids = record["article_ids"]
            release_date = record["release_date"]

            releases.append(GridRelease(self.dag_id, article_ids, pendulum.parse(release_date)))
        return releases

    def list_releases(self, **kwargs):
        """Lists all GRID releases for a given month and publishes their article_id's and
        release_date's as an XCom.

        :param kwargs: the context passed from the BranchPythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: the identifier of the task to execute next.
        """

        execution_date = kwargs["execution_date"]
        next_execution_date = kwargs["next_execution_date"]
        records = list_grid_records(execution_date, next_execution_date, GridTelescope.GRID_DATASET_URL)

        continue_dag = len(records)
        if continue_dag:
            # Push messages
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(GridTelescope.RELEASE_INFO, records, execution_date)

        return continue_dag

    def download(self, releases: List[GridRelease], **kwargs):
        """Task to download the GRID releases for a given month.

        :param releases: a list of GRID releases.
        :return: None.
        """

        # Download each release
        for release in releases:
            release.download()

    def upload_downloaded(self, releases: List[GridRelease], **kwargs):
        """Task to upload the downloaded GRID releases for a given month.

        :param releases: a list of GRID releases.
        :return: None.
        """

        # Upload each downloaded release
        for release in releases:
            upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, releases: List[GridRelease], **kwargs):
        """Task to extract the GRID releases for a given month.

        :param releases: a list of GRID releases.
        :return: None.
        """

        # Extract each release
        for release in releases:
            release.extract()

    def transform(self, releases: List[GridRelease], **kwargs):
        """Task to transform the GRID releases for a given month.

        :param releases: a list of GRID releases.
        :return: None.
        """

        # Transform each release
        for release in releases:
            release.transform()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Richard Hosking, James Diprose

# The keywords airflow and DAG are required to load the DAGs from this file, see bullet 2 in the Apache Airflow FAQ:
# https://airflow.apache.org/docs/stable/faq.html

"""
A DAG that produces the dois table and aggregated tables for the dashboards.

Each release is saved to the following BigQuery tables:
    <project_id>.observatory.countryYYYYMMDD
    <project_id>.observatory.doiYYYYMMDD
    <project_id>.observatory.funderYYYYMMDD
    <project_id>.observatory.groupYYYYMMDD
    <project_id>.observatory.institutionYYYYMMDD
    <project_id>.observatory.journalYYYYMMDD
    <project_id>.observatory.publisherYYYYMMDD
    <project_id>.observatory.regionYYYYMMDD
    <project_id>.observatory.subregionYYYYMMDD

Every week the following tables are overwritten for visualisation in the Data Studio dashboards:
    <project_id>.coki_dashboards.country
    <project_id>.coki_dashboards.doi
    <project_id>.coki_dashboards.funder
    <project_id>.coki_dashboards.group
    <project_id>.coki_dashboards.institution
    <project_id>.coki_dashboards.journal
    <project_id>.coki_dashboards.publisher
    <project_id>.coki_dashboards.region
    <project_id>.coki_dashboards.subregion
"""

from academic_observatory_workflows.workflows.doi_workflow import DoiWorkflow

# Outputs data into:
doi_workflow = DoiWorkflow()
globals()[doi_workflow.dag_id] = doi_workflow.make_dag()



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# Author: Aniek Roelofs

import gzip
import logging
import math
import os
import shutil
import subprocess
import tarfile
from concurrent.futures import as_completed, ProcessPoolExecutor
from io import BytesIO
from subprocess import Popen
from typing import List

import boto3
import jsonlines
import pendulum
import xmltodict
from airflow.exceptions import AirflowException, AirflowSkipException
from airflow.hooks.base_hook import BaseHook
from airflow.models.taskinstance import TaskInstance
from airflow.models.variable import Variable
from observatory.platform.utils.airflow_utils import AirflowConns
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.gc_utils import (
    aws_to_google_cloud_storage_transfer,
    storage_bucket_exists,
)
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.workflows.stream_telescope import (
    StreamRelease,
    StreamTelescope,
)

from academic_observatory_workflows.config import schema_folder as default_schema_folder


class OrcidRelease(StreamRelease):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
        first_release: bool,
        max_processes: int,
        batch_size: int = 500,
        log_count: int = 5000,
    ):
        """Construct an OrcidRelease instance

        :param dag_id: the id of the DAG.
        :param start_date: the start_date of the release.
        :param end_date: the end_date of the release.
        :param first_release: whether this is the first release that is processed for this DAG.
        :param max_processes: Max processes used for transforming files.
        :param batch_size: the size of batches used when transforming files.
        :param log_count: after how many iterations to print transform log update.
        """
        download_files_regex = r".*.xml$"
        transform_files_regex = r".*.jsonl.gz"
        super().__init__(
            dag_id,
            start_date,
            end_date,
            first_release,
            download_files_regex=download_files_regex,
            transform_files_regex=transform_files_regex,
        )
        self.max_processes = max_processes
        self.batch_size = batch_size
        self.log_count = log_count

    @property
    def modified_records_path(self) -> str:
        """Get the path to the file with ids of modified records.

        :return: the file path.
        """
        return os.path.join(self.download_folder, "modified_records.txt")

    def transfer(self, max_retries):
        """Sync files from AWS bucket to Google Cloud bucket

        :param max_retries: Number of max retries to try the transfer
        :return: None.
        """
        aws_access_key_id, aws_secret_access_key = get_aws_conn_info()

        gc_download_bucket = Variable.get(AirflowVars.ORCID_BUCKET)
        gc_project_id = Variable.get(AirflowVars.PROJECT_ID)
        last_modified_since = None if self.first_release else self.start_date

        success = False
        total_count = 0

        for i in range(max_retries):
            if success:
                break
            success, objects_count = aws_to_google_cloud_storage_transfer(
                aws_access_key_id,
                aws_secret_access_key,
                aws_bucket=OrcidTelescope.SUMMARIES_BUCKET,
                include_prefixes=[],
                gc_project_id=gc_project_id,
                gc_bucket=gc_download_bucket,
                description="Transfer ORCID data from airflow telescope",
                last_modified_since=last_modified_since,
            )
            total_count += objects_count

        if not success:
            raise AirflowException(f"Google Storage Transfer unsuccessful, status: {success}")

        logging.info(f"Total number of objects transferred: {total_count}")
        if total_count == 0:
            raise AirflowSkipException("No objects to transfer")

    def download_transferred(self):
        """Download the updated records from the Google Cloud bucket to a local directory using gsutil.
        If the run processes the first release it will download all files. If it is a later release, it will check
        the ORCID lambda file which tracks which records are modified. Only the modified records will be downloaded.

        :return: None.
        """
        aws_access_key_id, aws_secret_access_key = get_aws_conn_info()

        gc_download_bucket = Variable.get(AirflowVars.ORCID_BUCKET)

        # Authenticate gcloud with service account
        args = [
            "gcloud",
            "auth",
            "activate-service-account",
            f"--key-file" f"={os.environ['GOOGLE_APPLICATION_CREDENTIALS']}",
        ]
        proc: Popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        run_subprocess_cmd(proc, args)

        logging.info(f"Downloading transferred files from Google Cloud bucket: {gc_download_bucket}")
        log_path = os.path.join(self.download_folder, "cp.log")
        if self.first_release:
            # Download all records from bucket
            args = [
                "gsutil",
                "-m",
                "-q",
                "cp",
                "-L",
                log_path,
                "-r",
                f"gs://{gc_download_bucket}",
                self.download_folder,
            ]
            proc: Popen = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        else:
            # Download only modified records from bucket
            write_modified_record_blobs(
                self.start_date,
                self.end_date,
                aws_access_key_id,
                aws_secret_access_key,
                gc_download_bucket,
                self.modified_records_path,
            )
            args = ["gsutil", "-m", "-q", "cp", "-L", log_path, "-I", self.download_folder]
            proc: Popen = subprocess.Popen(
                args, stdin=open(self.modified_records_path), stdout=subprocess.PIPE, stderr=subprocess.PIPE
            )
        run_subprocess_cmd(proc, args)

    def transform(self):
        """Transform the ORCID records in parallel.
        Each file is 1 record, after the single file is transformed the data is appended to a .jsonl.gz file

        :return: None.
        """
        logging.info(f"Using {self.max_processes} workers for multithreading")
        count = 0
        files_to_process = self.download_files
        num_batches = math.ceil(len(files_to_process) / self.batch_size)

        # Process the files in batches because there are so many and we don't get feedback when
        for b in range(num_batches):
            logging.info(f"Transforming batch: {b}")
            index = b * self.batch_size
            batch_files = files_to_process[index : index + self.batch_size]

            with ProcessPoolExecutor(max_workers=self.max_processes) as executor:
                futures = list()
                for file_path in batch_files:
                    future = executor.submit(transform_single_file, file_path, self.transform_folder)
                    futures.append(future)

                for future in as_completed(futures):
                    future.result()
                    count += 1
                    if count % self.log_count == 0:
                        logging.info(f"Transformed {count} files")

        # Loop through directories with individual files, concatenate files in each directory into 1 gzipped file.
        logging.info("Finished transforming individual files, concatenating & compressing files")
        for root, dirs, files in os.walk(self.transform_folder):
            if root == self.transform_folder:
                continue
            file_dir = os.path.basename(root)
            transform_path = os.path.join(self.transform_folder, file_dir + ".jsonl.gz")
            with gzip.GzipFile(transform_path, mode="wb") as f_out:
                for name in files:
                    with open(os.path.join(root, name), "rb") as f_in:
                        shutil.copyfileobj(f_in, f_out)


class OrcidTelescope(StreamTelescope):
    """ORCID telescope"""

    DAG_ID = "orcid"

    SUMMARIES_BUCKET = "v2.0-summaries"
    LAMBDA_BUCKET = "orcid-lambda-file"
    LAMBDA_OBJECT = "last_modified.csv.tar"
    S3_HOST = "s3.eu-west-1.amazonaws.com"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2018, 5, 14),
        schedule_interval: str = "@weekly",
        dataset_id: str = "orcid",
        dataset_description: str = "",
        table_descriptions: dict = None,
        queue: str = "remote_queue",
        merge_partition_field: str = "orcid_identifier.uri",
        bq_merge_days: int = 7,
        schema_folder: str = default_schema_folder(),
        batch_load: bool = True,
        airflow_vars: List = None,
        airflow_conns: List = None,
        max_processes: int = min(32, os.cpu_count() + 4),
    ):
        """Construct an OrcidTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the dataset id.
        :param dataset_description: the dataset description.
        :param queue: the queue that the telescope should run on.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions.
        :param merge_partition_field: the BigQuery field used to match partitions for a merge
        :param bq_merge_days: how often partitions should be merged (every x days)
        :param schema_folder: the SQL schema path.
        :param batch_load: whether all files in the transform folder are loaded into 1 table at once
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param airflow_conns: list of airflow connection keys, for each connection it is checked if it exists in airflow
        :param max_processes: Max processes used for parallel transforming.
        """
        if table_descriptions is None:
            table_descriptions = {
                dag_id: "The ORCID (Open Researcher and Contributor ID) is a nonproprietary "
                "alphanumeric code to uniquely identify authors and contributors of "
                "scholarly communication, see: https://orcid.org/."
            }

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
                AirflowVars.ORCID_BUCKET,
            ]
        if airflow_conns is None:
            airflow_conns = [AirflowConns.ORCID]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            merge_partition_field,
            bq_merge_days,
            schema_folder,
            dataset_description=dataset_description,
            table_descriptions=table_descriptions,
            queue=queue,
            airflow_vars=airflow_vars,
            airflow_conns=airflow_conns,
            batch_load=batch_load,
        )
        self.max_processes = max_processes

        self.add_setup_task_chain([self.check_dependencies, self.get_release_info])
        self.add_task_chain(
            [self.transfer, self.download_transferred, self.transform, self.upload_transformed, self.bq_load_partition]
        )
        self.add_task_chain([self.bq_delete_old, self.bq_append_new, self.cleanup], trigger_rule="none_failed")

    def make_release(self, **kwargs) -> OrcidRelease:
        """Make a release instance. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are
        passed to this argument.
        :return: an OrcidRelease instance.
        """
        ti: TaskInstance = kwargs["ti"]
        start_date, end_date, first_release = ti.xcom_pull(key=OrcidTelescope.RELEASE_INFO, include_prior_dates=True)

        release = OrcidRelease(
            self.dag_id, pendulum.parse(start_date), pendulum.parse(end_date), first_release, self.max_processes
        )
        return release

    def check_dependencies(self, **kwargs) -> bool:
        """Check dependencies of DAG. Add to parent method to additionally check whether the Google Cloud bucket
        that is used to sync ORCID data exists.

        :return: True if dependencies are valid.
        """
        super().check_dependencies()

        orcid_bucket_name = Variable.get(AirflowVars.ORCID_BUCKET)
        if not storage_bucket_exists(orcid_bucket_name):
            raise AirflowException(f"Bucket to store ORCID download data does not exist ({orcid_bucket_name})")
        return True

    def transfer(self, release: OrcidRelease, **kwargs):
        """Task to transfer data of the ORCID release.

        :param release: an OrcidRelease instance.
        :return: None.
        """
        release.transfer(self.max_retries)

    def download_transferred(self, release: OrcidRelease, **kwargs):
        """Task to download the transferred data of the ORCID release.

        :param release: an OrcidRelease instance.
        :return: None.
        """
        release.download_transferred()

    def transform(self, release: OrcidRelease, **kwargs):
        """Task to transform data of the ORCID release.

        :param release: an OrcidRelease instance.
        :return: None.
        """
        release.transform()


def get_aws_conn_info() -> (str, str):
    """Get the AWS access key id and secret access key from the ORCID airflow connection.

    :return: access key id and secret access key
    """
    conn = BaseHook.get_connection(AirflowConns.ORCID)
    access_key_id = conn.login
    secret_access_key = conn.password

    return access_key_id, secret_access_key


def transform_single_file(download_path: str, transform_folder: str):
    """ Transform a single ORCID file/record.
    The xml file is turned into a dictionary, a record should have either a valid 'record' section or an 'error'
    section. The keys of the dictionary are slightly changed so they are valid BigQuery fields.
    The dictionary is appended to a jsonl file
    :param download_path: The path to the file with the ORCID record.
    :param transform_folder: The path where transformed files will be saved.
    :return: None.
    """

    file_name = os.path.basename(download_path)
    file_dir = os.path.join(transform_folder, file_name[-7:-4])  # last three digits are used for subdir

    # Create subdirectory if it does not exist yet, even with if statement it will still raise FileExistsError
    # sometimes
    if not os.path.exists(file_dir):
        try:
            os.mkdir(file_dir)
        except FileExistsError:
            pass

    transform_path = os.path.join(file_dir, os.path.splitext(file_name)[0] + ".jsonl")
    # Skip if file already exists
    if os.path.exists(transform_path):
        return

    # Create dict of data from summary xml file
    with open(download_path, "r") as f:
        orcid_dict = xmltodict.parse(f.read())

    # Get record
    orcid_record = orcid_dict.get("record:record")

    # Some records do not have a 'record', but only 'error', this will be stored in the BQ table.
    if not orcid_record:
        orcid_record = orcid_dict.get("error:error")
    if not orcid_record:
        raise AirflowException(f"Key error for file: {download_path}")

    orcid_record = change_keys(orcid_record, convert)
    with jsonlines.open(transform_path, "w") as writer:
        writer.write(orcid_record)

    del orcid_dict
    del orcid_record


def run_subprocess_cmd(proc: Popen, args: list):
    """Execute and wait for subprocess to finish, also handle stdout & stderr from process.

    :param proc: subprocess proc
    :param args: args list that was passed on to subprocess
    :return: None.
    """
    logging.info(f"Executing bash command: {subprocess.list2cmdline(args)}")
    out, err = wait_for_process(proc)
    if out:
        logging.info(out)
    if err:
        logging.info(err)
    if proc.returncode != 0:
        # Don't raise exception if the only error is because blobs could not be found in bucket
        err_lines = err.split("\n")
        for line in err_lines[:]:
            if not line or "CommandException: No URLs matched:" in line or "could not be transferred." in line:
                err_lines.remove(line)
        if err_lines:
            raise AirflowException("bash command failed")
    logging.info("Finished cmd successfully")


def write_modified_record_blobs(
    start_date: pendulum.DateTime,
    end_date: pendulum.DateTime,
    aws_access_key_id: str,
    aws_secret_access_key: str,
    gc_download_bucket: str,
    modified_records_path: str,
) -> int:
    """Download the ORCID lambda file (last_modified.csv.tar) from AWS and use file to write the full Google Cloud
    blob names of modified records.
    The tar file is opened in memory and contains the ORCID record IDs, sorted by last modified date.

    :param start_date: Start date of the release
    :param end_date: End date of the release
    :param aws_access_key_id: AWS access key id
    :param aws_secret_access_key: AWS secret access key
    :param gc_download_bucket: Name of Google Cloud bucket with ORCID records
    :param modified_records_path: Path to file with the blob names of modified records
    :return: The number of modified records.
    """
    logging.info(f"Writing modified records to {modified_records_path}")

    # orcid lambda file, containing info on last_modified dates of records
    aws_lambda_bucket = OrcidTelescope.LAMBDA_BUCKET
    aws_lambda_object = OrcidTelescope.LAMBDA_OBJECT

    s3client = boto3.client("s3", aws_access_key_id=aws_access_key_id, aws_secret_access_key=aws_secret_access_key)
    lambda_obj = s3client.get_object(Bucket=aws_lambda_bucket, Key=aws_lambda_object)
    lambda_content = lambda_obj["Body"].read()

    modified_records_count = 0
    # open tar file in memory
    with tarfile.open(fileobj=BytesIO(lambda_content)) as tar, open(modified_records_path, "w") as f:
        for tar_resource in tar:
            if tar_resource.isfile():
                # extract last modified file in memory
                inner_file_bytes = tar.extractfile(tar_resource).read().decode().split("\n")
                for line in inner_file_bytes[1:]:
                    elements = line.split(",")
                    orcid_record = elements[0]

                    # parse through line by line, check if last_modified timestamp is between start/end date
                    last_modified_date = pendulum.parse(elements[3])

                    # skip records that are too new, not included in this release
                    if last_modified_date > end_date:
                        continue
                    # use records between start date and end date
                    elif last_modified_date >= start_date:
                        directory = orcid_record[-3:]
                        f.write(f"gs://{gc_download_bucket}/{directory}/{orcid_record}.xml" + "\n")
                        modified_records_count += 1
                    # stop when reached records before start date, not included in this release
                    else:
                        break

    return modified_records_count


def convert(k: str) -> str:
    """Convert key of dictionary to valid BQ key.

    :param k: Key
    :return: The converted key
    """
    if len(k.split(":")) > 1:
        k = k.split(":")[1]
    if k.startswith("@") or k.startswith("#"):
        k = k[1:]
    k = k.replace("-", "_")
    return k


def change_keys(obj, convert):
    """Recursively goes through the dictionary obj and replaces keys with the convert function.

    :param obj: The dictionary value, can be object of any type
    :param convert: The convert function.
    :return: The transformed object.
    """
    if isinstance(obj, (str, int, float)):
        return obj
    if isinstance(obj, dict):
        new = obj.__class__()
        for k, v in list(obj.items()):
            if k.startswith("@xmlns"):
                pass
            else:
                new[convert(k)] = change_keys(v, convert)
    elif isinstance(obj, (list, set, tuple)):
        new = obj.__class__(change_keys(v, convert) for v in obj)
    else:
        return obj
    return new



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs

from __future__ import annotations

import logging
import os
import re
from concurrent.futures import ThreadPoolExecutor, as_completed, ProcessPoolExecutor
from typing import List, Tuple, Union

import jsonlines
import pendulum
import requests
from airflow.exceptions import AirflowSkipException
from airflow.models.taskinstance import TaskInstance
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.url_utils import get_user_agent
from observatory.platform.utils.workflow_utils import upload_files_from_list
from observatory.platform.workflows.stream_telescope import (
    StreamRelease,
    StreamTelescope,
)
from tenacity import RetryError, retry, stop_after_attempt, wait_exponential, wait_fixed

from academic_observatory_workflows.config import schema_folder as default_schema_folder


class CrossrefEventsRelease(StreamRelease):
    def __init__(
        self,
        dag_id: str,
        start_date: pendulum.DateTime,
        end_date: pendulum.DateTime,
        first_release: bool,
        mailto: str,
        max_threads: int,
        max_processes: int,
    ):
        """Construct a CrossrefEventsRelease instance

        :param dag_id: the id of the DAG.
        :param start_date: the start_date of the release.
        :param end_date: the end_date of the release.
        :param first_release: whether this is the first release that is processed for this DAG
        :param mailto: Email address used in the download url
        :param max_threads: Max threads used for parallel downloading
        :param max_processes: max processes for transforming files.
        """
        download_files_regex = r".*.jsonl$"
        transform_files_regex = r".*.jsonl$"
        super().__init__(
            dag_id,
            start_date,
            end_date,
            first_release,
            download_files_regex=download_files_regex,
            transform_files_regex=transform_files_regex,
        )
        self.mailto = mailto
        self.max_threads = max_threads
        self.max_processes = max_processes

    @property
    def urls(self) -> list:
        urls = []
        start_date = self.start_date.date()
        end_date = self.end_date.date()
        period = pendulum.period(start_date, end_date)
        for dt in period.range("days"):
            date_str = dt.strftime("%Y-%m-%d")
            start_date = date_str
            end_date = date_str

            events_url = (
                f"https://api.eventdata.crossref.org/v1/events?mailto={self.mailto}"
                f"&from-collected-date={start_date}&until-collected-date={end_date}&rows=1000"
            )
            edited_url = (
                f"https://api.eventdata.crossref.org/v1/events/edited?mailto={self.mailto}"
                f"&from-updated-date={start_date}&until-updated-date={end_date}&rows=1000"
            )
            deleted_url = (
                f"https://api.eventdata.crossref.org/v1/events/deleted?mailto={self.mailto}"
                f"&from-updated-date={start_date}&until-updated-date={end_date}&rows=1000"
            )

            urls.append(events_url)
            if not self.first_release:
                urls.append(edited_url)
                urls.append(deleted_url)
        return urls

    def batch_path(self, url, cursor: bool = False) -> str:
        """Gets the appropriate file path for a single batch, either for an events or cursor file.

        :param url: The url used for a specific batch
        :param cursor: Whether this is a cursor file or file with actual events
        :return: Path to the events or cursor file
        """
        event_type, date = parse_event_url(url)
        if cursor:
            return os.path.join(self.download_folder, f"{event_type}_{date}_cursor.txt")
        else:
            return os.path.join(self.download_folder, f"{event_type}_{date}.jsonl")

    def download(self):
        """Download all events.

        :return: None.
        """
        logging.info(f"Downloading events, no. workers: {self.max_threads}")
        logging.info(f"Downloading using these URLs, but with different start and end dates: {self.urls[0]}")

        with ThreadPoolExecutor(max_workers=self.max_threads) as executor:
            futures = []
            for i, url in enumerate(self.urls):
                futures.append(executor.submit(self.download_batch, i, url))
            for future in as_completed(futures):
                future.result()
        if len(self.download_files) == 0:
            raise AirflowSkipException("No events found")

    def download_batch(self, i: int, url: str):
        """Download one day of events. When the download finished successfully, the generated cursor file is deleted.
        If there is a cursor file available at the start, it means that a previous download attempt failed. If there
        is an events file available and no cursor file, it means that a previous download attempt was successful,
        so these events will not be downloaded again.

        :param i: URL counter
        :param url: The url from which to download events
        :return: None.
        """
        events_path = self.batch_path(url)
        cursor_path = self.batch_path(url, cursor=True)
        event_type, date = parse_event_url(url)

        # if events file exists but no cursor file, previous request has finished & successful
        if os.path.isfile(events_path) and not os.path.isfile(cursor_path):
            logging.info(f"{i + 1}.{event_type} Skipped, already finished: {date}")
            return

        logging.info(f"{i + 1}.{event_type} Downloading date: {date}")
        headers = {"User-Agent": get_user_agent(package_name="academic_observatory_workflows")}
        next_cursor, counts, total_events = download_events(url, headers, events_path, cursor_path)
        counter = counts
        while next_cursor:
            tmp_url = url + f"&cursor={next_cursor}"
            next_cursor, counts, _ = download_events(tmp_url, headers, events_path, cursor_path)
            counter += counts

        if os.path.isfile(cursor_path):
            os.remove(cursor_path)
        logging.info(
            f"{i + 1}.{event_type} successful, date: {date}, total no. events: {total_events}, downloaded "
            f"events: {counter}"
        )

    def transform(self):
        """Transform all events.

        :return: None.
        """
        logging.info(f"Transforming events, no. workers: {self.max_processes}")

        with ProcessPoolExecutor(max_workers=self.max_processes) as executor:
            futures = []
            for file in self.download_files:
                futures.append(executor.submit(transform_batch, file, self.transform_folder))
            for future in as_completed(futures):
                future.result()


def transform_batch(download_path: str, transform_folder: str):
    """Transform one day of events.

    :param download_path: The path to the downloaded file.
    :param transform_folder: the transform folder.
    :return: None.
    """

    file_name = os.path.basename(download_path)
    transform_path = os.path.join(transform_folder, file_name)

    logging.info(f"Transforming file: {download_path}")
    logging.info(f"Saving to: {transform_path}")
    with jsonlines.open(download_path, "r") as reader:
        with jsonlines.open(transform_path, "w") as writer:
            for event in reader:
                event = transform_events(event)
                writer.write(event)

    logging.info(f"Finished: {file_name}")


class CrossrefEventsTelescope(StreamTelescope):
    """Crossref Events telescope"""

    DAG_ID = "crossref_events"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2018, 5, 14),
        schedule_interval: str = "@weekly",
        dataset_id: str = "crossref",
        dataset_description: str = "The Crossref Events dataset: https://www.eventdata.crossref.org/guide/",
        queue: str = "remote_queue",
        merge_partition_field: str = "id",
        bq_merge_days: int = 7,
        schema_folder: str = default_schema_folder(),
        batch_load: bool = True,
        airflow_vars: List = None,
        mailto: str = "aniek.roelofs@curtin.edu.au",
        max_threads: int = min(32, os.cpu_count() + 4),
        max_processes: int = os.cpu_count(),
    ):
        """Construct a CrossrefEventsTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the dataset id.
        :param dataset_description: the dataset description.
        :param queue: the queue that the tasks should run on.
        :param merge_partition_field: the BigQuery field used to match partitions for a merge
        :param bq_merge_days: how often partitions should be merged (every x days)
        :param schema_folder: the SQL schema path.
        :param batch_load: whether all files in the transform folder are loaded into 1 table at once
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow
        :param mailto: Email address used in the download url
        :param max_threads: Max processes used for parallel downloading, default is based on 7 days x 3 url categories
        :param max_processes: max processes for transforming files.
        """

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            merge_partition_field,
            bq_merge_days,
            schema_folder,
            dataset_description=dataset_description,
            queue=queue,
            batch_load=batch_load,
            airflow_vars=airflow_vars,
        )
        self.mailto = mailto
        self.max_threads = max_threads
        self.max_processes = max_processes

        self.add_setup_task_chain([self.check_dependencies, self.get_release_info])
        self.add_task_chain(
            [self.download, self.upload_downloaded, self.transform, self.upload_transformed, self.bq_load_partition]
        )
        self.add_task_chain([self.bq_delete_old, self.bq_append_new, self.cleanup], trigger_rule="none_failed")

    def make_release(self, **kwargs) -> CrossrefEventsRelease:
        """Make a Release instance

        :param kwargs: The context passed from the PythonOperator.
        :return: CrossrefEventsRelease
        """
        ti: TaskInstance = kwargs["ti"]
        start_date, end_date, first_release = ti.xcom_pull(
            key=CrossrefEventsTelescope.RELEASE_INFO, include_prior_dates=True
        )

        start_date = pendulum.parse(start_date)
        end_date = pendulum.parse(end_date)

        release = CrossrefEventsRelease(
            self.dag_id, start_date, end_date, first_release, self.mailto, self.max_threads, self.max_processes
        )
        return release

    def download(self, release: CrossrefEventsRelease, **kwargs):
        """Task to download the CrossrefEventsRelease release.

        :param release: a CrossrefEventsRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.download()

    def upload_downloaded(self, release: CrossrefEventsRelease, **kwargs):
        """Upload the downloaded files for the given release.

        :param release: a CrossrefEventsRelease instance
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        upload_files_from_list(release.download_files, release.download_bucket)

    def transform(self, release: CrossrefEventsRelease, **kwargs):
        """Task to transform the CrossrefEventsRelease release.

        :param release: a CrossrefEventsRelease instance.
        :param kwargs: The context passed from the PythonOperator.
        :return: None.
        """
        release.transform()


@retry(
    stop=stop_after_attempt(3), wait=wait_fixed(20) + wait_exponential(multiplier=10, exp_base=3, max=60 * 10),
)
def get_response(url: str, headers: dict):
    """Get response from the url with given headers and retry for certain status codes.

    :param url: The url
    :param headers: The headers dict
    :return: The response
    """
    response = requests.get(url, headers=headers)
    if response.status_code in [500, 400, 429]:
        logging.info(
            f'Downloading events from url: {url}, attempt: {get_response.retry.statistics["attempt_number"]}, '
            f'idle for: {get_response.retry.statistics["idle_for"]}'
        )
        raise ConnectionError("Retrying url")
    return response


def parse_event_url(url: str) -> (str, str):
    """Parse the URL to get the event type and date

    :param url: The url
    :return: The event type and date
    """
    event_type = url.split("?mailto")[0].split("/")[-1]
    if event_type == "events":
        date = url.split("from-collected-date=")[1].split("&")[0]
    else:
        date = url.split("from-updated-date=")[1].split("&")[0]

    return event_type, date


def download_events(url: str, headers: dict, events_path: str, cursor_path: str) -> Tuple[Union[str, None], int, int]:
    """Extract the events from the given url until no new cursor is returned or a RetryError occurs.
    The extracted events are appended to a jsonl file and the cursors are written to a text file.

    :param url: The url
    :param headers: The headers dict
    :param events_path: Path to the file in which events are stored.
    :param cursor_path: Path to the file where cursors are stored.
    :return: next_cursor, counter of events and total number of events according to the response
    """
    try:
        response = get_response(url, headers)
    except RetryError:
        # Try again with rows set to 100
        url = re.sub("rows=[0-9]*", "rows=100", url)
        response = get_response(url, headers)

    if response.status_code == 200:
        response_json = response.json()
        total_events = response_json["message"]["total-results"]
        events = response_json["message"]["events"]
        next_cursor = response_json["message"]["next-cursor"]
        counter = len(events)

        # append events and cursor
        if events:
            with open(events_path, "a") as f:
                with jsonlines.Writer(f) as writer:
                    writer.write_all(events)
        if next_cursor:
            with open(cursor_path, "a") as f:
                f.write(next_cursor + "\n")
        return next_cursor, counter, total_events
    else:
        raise ConnectionError(f"Error requesting url: {url}, response: {response.text}")


def transform_events(event):
    """Transform the dictionary with event data by replacing '-' with '_' in key names, converting all int values to
    string except for the 'total' field and parsing datetime columns for a valid datetime.

    :param event: The event dictionary
    :return: The updated event dictionary
    """
    if isinstance(event, (str, int, float)):
        return event
    if isinstance(event, dict):
        new = event.__class__()
        for k, v in event.items():
            if isinstance(v, int) and k != "total":
                v = str(v)
            if k in ["timestamp", "occurred_at", "issued", "dateModified", "updated_date"]:
                try:
                    v = str(pendulum.parse(v))
                except ValueError:
                    v = "0001-01-01T00:00:00Z"

            # Replace hyphens with underscores for BigQuery compatibility
            k = k.replace("-", "_")

            # Replace @ symbol in keys left by DataCite between the 15 and 22 March 2019
            k = k.replace("@", "")

            new[k] = transform_events(v)
        return new



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, Richard Hosking, James Diprose

from __future__ import annotations

import gzip
import io
import json
import logging
import os
import random
import shutil
import subprocess
import xml.etree.ElementTree as ET
from typing import Dict, List, Tuple

import jsonlines
import pendulum
import requests
from airflow.api.common.experimental.pool import create_pool
from airflow.exceptions import AirflowException
from airflow.models.taskinstance import TaskInstance

from academic_observatory_workflows.config import schema_folder as default_schema_folder
from airflow.models.variable import Variable
from observatory.platform.utils.airflow_utils import AirflowVars
from observatory.platform.utils.gc_utils import (
    bigquery_sharded_table_id,
    bigquery_table_exists,
)
from observatory.platform.utils.proc_utils import wait_for_process
from observatory.platform.utils.url_utils import retry_session
from observatory.platform.utils.workflow_utils import upload_files_from_list
from observatory.platform.workflows.snapshot_telescope import (
    SnapshotRelease,
    SnapshotTelescope,
)


class CrossrefFundrefRelease(SnapshotRelease):
    def __init__(self, dag_id: str, release_date: pendulum.datetime, url: str):
        """Construct a CrossrefFundrefRelease

        :param dag_id: The DAG id.
        :param release_date: The release date.
        :param url: The url corresponding with this release date.
        """
        self.url = url

        download_files_regex = "crossref_fundref.tar.gz"
        extract_files_regex = "crossref_fundref.rdf"
        transform_files_regex = "crossref_fundref.jsonl.gz"
        super().__init__(dag_id, release_date, download_files_regex, extract_files_regex, transform_files_regex)

    @property
    def download_path(self) -> str:
        """Get the path to the downloaded file.

        :return: the file path.
        """
        return os.path.join(self.download_folder, "crossref_fundref.tar.gz")

    @property
    def extract_path(self) -> str:
        """Get the path to the extracted file.

        :return: the file path.
        """

        return os.path.join(self.extract_folder, "crossref_fundref.rdf")

    @property
    def transform_path(self) -> str:
        """Get the path to the transformed file.

        :return: the file path.
        """

        return os.path.join(self.transform_folder, "crossref_fundref.jsonl.gz")

    def download(self):
        """Downloads release tar.gz file from url."""

        logging.info(f"Downloading file: {self.download_path}, url: {self.url}")

        # A selection of headers to prevent 403/forbidden error.
        headers_list = [
            {
                "authority": "gitlab.com",
                "upgrade-insecure-requests": "1",
                "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) "
                "Chrome/83.0.4103.116 Safari/537.36",
                "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,"
                "*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
                "sec-fetch-site": "none",
                "sec-fetch-mode": "navigate",
                "sec-fetch-dest": "document",
                "accept-language": "en-GB,en-US;q=0.9,en;q=0.8",
            },
            {
                "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
                "Accept": "*/*",
                "Accept-Language": "en-US,en;q=0.5",
                "Referer": "https://gitlab.com/",
            },
        ]

        # Download release
        with requests.get(self.url, headers=random.choice(headers_list), stream=True) as response:
            with open(self.download_path, "wb") as file:
                shutil.copyfileobj(response.raw, file)

    def extract(self):
        """Extract release from gzipped tar file."""
        logging.info(f"Extracting file: {self.download_path}")
        # Tar file contains both README.md and registry.rdf, use tar -ztf to get path of 'registry.rdf'
        # Use this path to extract only registry.rdf to a new file.
        cmd = (
            f"registry_path=$(tar -ztf {self.download_path} | grep -m1 '/registry.rdf'); "
            f"tar -xOzf {self.download_path} $registry_path > {self.extract_path}"
        )
        p = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, executable="/bin/bash")
        stdout, stderr = wait_for_process(p)

        if stdout:
            logging.info(stdout)

        if stderr:
            raise AirflowException(f"bash command failed for {self.url}: {stderr}")

        logging.info(f"File extracted to: {self.extract_path}")

    def transform(self):
        """Transforms release by storing file content in gzipped json format. Relationships between funders are added.

        :return: None
        """

        # Strip leading whitespace from first line if present.
        strip_whitespace(self.extract_path)

        # Parse RDF funders data
        funders, funders_by_key = parse_fundref_registry_rdf(self.extract_path)
        funders = add_funders_relationships(funders, funders_by_key)

        # Transform FundRef release into JSON Lines format saving in memory buffer
        # Save in memory buffer to gzipped file
        with io.BytesIO() as bytes_io:
            with gzip.GzipFile(fileobj=bytes_io, mode="w") as gzip_file:
                with jsonlines.Writer(gzip_file) as writer:
                    writer.write_all(funders)

            with open(self.transform_path, "wb") as jsonl_gzip_file:
                jsonl_gzip_file.write(bytes_io.getvalue())

        logging.info(f"Success transforming release: {self.url}")


class CrossrefFundrefTelescope(SnapshotTelescope):
    """Crossref Fundref Telescope."""

    DAG_ID = "crossref_fundref"
    DATASET_ID = "crossref"
    RELEASES_URL = "https://gitlab.com/api/v4/projects/crossref%2Fopen_funder_registry/releases"

    def __init__(
        self,
        dag_id: str = DAG_ID,
        start_date: pendulum.DateTime = pendulum.datetime(2014, 3, 1),
        schedule_interval: str = "@weekly",
        dataset_id: str = DATASET_ID,
        schema_folder: str = default_schema_folder(),
        table_descriptions: Dict = None,
        catchup: bool = True,
        airflow_vars: List = None,
    ):

        """Construct a CrossrefFundrefTelescope instance.

        :param dag_id: the id of the DAG.
        :param start_date: the start date of the DAG.
        :param schedule_interval: the schedule interval of the DAG.
        :param dataset_id: the BigQuery dataset id.
        :param schema_folder: the SQL schema path.
        :param table_descriptions: a dictionary with table ids and corresponding table descriptions.
        :param catchup: whether to catchup the DAG or not.
        :param airflow_vars: list of airflow variable keys, for each variable it is checked if it exists in airflow.
        """

        if table_descriptions is None:
            table_descriptions = {
                dag_id: "The Funder Registry dataset: " "https://www.crossref.org/services/funder-registry/"
            }

        if airflow_vars is None:
            airflow_vars = [
                AirflowVars.DATA_PATH,
                AirflowVars.PROJECT_ID,
                AirflowVars.DATA_LOCATION,
                AirflowVars.DOWNLOAD_BUCKET,
                AirflowVars.TRANSFORM_BUCKET,
            ]
        super().__init__(
            dag_id,
            start_date,
            schedule_interval,
            dataset_id,
            schema_folder,
            table_descriptions=table_descriptions,
            catchup=catchup,
            airflow_vars=airflow_vars,
        )

        # Create Gitlab pool to limit the number of connections to Gitlab, which is very quick to block requests if
        # there are too many at once.
        pool_name = "gitlab_pool"
        num_slots = 2
        description = "A pool to limit the connections to Gitlab."
        create_pool(pool_name, num_slots, description)

        self.add_setup_task(self.check_dependencies)
        self.add_setup_task(self.get_release_info, pool=pool_name)
        self.add_task(self.download, pool=pool_name)
        self.add_task(self.upload_downloaded)
        self.add_task(self.extract)
        self.add_task(self.transform)
        self.add_task(self.upload_transformed)
        self.add_task(self.bq_load)
        self.add_task(self.cleanup)

    def make_release(self, **kwargs) -> List[CrossrefFundrefRelease]:
        """Make release instances. The release is passed as an argument to the function (TelescopeFunction) that is
        called in 'task_callable'.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html for a list of the keyword arguments that are
        passed to this argument.
        :return: a list of GeonamesRelease instances.
        """

        ti: TaskInstance = kwargs["ti"]
        release_info = ti.xcom_pull(
            key=CrossrefFundrefTelescope.RELEASE_INFO,
            task_ids=self.get_release_info.__name__,
            include_prior_dates=False,
        )
        releases = []
        for release in release_info:
            release_date = pendulum.parse(release["date"])
            releases.append(CrossrefFundrefRelease(self.dag_id, release_date, release["url"]))
        return releases

    def get_release_info(self, **kwargs) -> bool:
        """Based on a list of all releases, checks which ones were released between the prev and this execution date
        of the DAG. If the release falls within the time period mentioned above, checks if a bigquery table doesn't
        exist yet for the release. A list of releases that passed both checks is passed to the next tasks. If the
        list is empty the workflow will stop.

        :param kwargs: the context passed from the PythonOperator. See
        https://airflow.apache.org/docs/stable/macros-ref.html
        for a list of the keyword arguments that are passed to this argument.
        :return: None.
        """

        # Get variables
        project_id = Variable.get(AirflowVars.PROJECT_ID)

        # List releases between a start date and an end date
        prev_execution_date = pendulum.instance(kwargs["prev_execution_date"])
        execution_date = pendulum.instance(kwargs["execution_date"])
        releases_list = list_releases(prev_execution_date, execution_date)
        logging.info(f"Releases between prev ({prev_execution_date}) and current ({execution_date}) execution date:")
        logging.info(releases_list)

        # Check if the BigQuery table for each release already exists and only process release if the table
        # doesn't exist
        releases_list_out = []
        for release in releases_list:
            table_id = bigquery_sharded_table_id(CrossrefFundrefTelescope.DAG_ID, release["date"])
            logging.info("Checking if bigquery table already exists:")
            if bigquery_table_exists(project_id, self.dataset_id, table_id):
                logging.info(
                    f"Skipping as table exists for {release['url']}: " f"{project_id}.{self.dataset_id}.{table_id}"
                )
            else:
                logging.info(f"Table does not exist yet, processing {release['url']} in this workflow")
                release["date"] = release["date"].format("YYYYMMDD")
                releases_list_out.append(release)

        # If releases_list_out contains items then the DAG will continue (return True) otherwise it will
        # stop (return False)
        continue_dag = len(releases_list_out) > 0
        if continue_dag:
            ti: TaskInstance = kwargs["ti"]
            ti.xcom_push(CrossrefFundrefTelescope.RELEASE_INFO, releases_list_out, execution_date)
        return continue_dag

    def download(self, releases: List[CrossrefFundrefRelease], **kwargs):
        """Task to download the releases.

        :param releases: a list with Crossref Fundref releases.
        :return: None.
        """
        for release in releases:
            release.download()

    def upload_downloaded(self, releases: List[CrossrefFundrefRelease], **kwargs):
        """Task to upload the downloaded releases.

        :param releases: a list with Crossref Fundref releases.
        :return: None.
        """
        for release in releases:
            upload_files_from_list(release.download_files, release.download_bucket)

    def extract(self, releases: List[CrossrefFundrefRelease], **kwargs):
        """Task to extract the releases.

        :param releases: a list with Crossref Fundref releases.
        :return: None.
        """
        for release in releases:
            release.extract()

    def transform(self, releases: List[CrossrefFundrefRelease], **kwargs):
        """Task to transform the releases.

        :param releases: a list with Crossref Fundref releases.
        :return: None.
        """
        for release in releases:
            release.transform()


def list_releases(start_date: pendulum.DateTime, end_date: pendulum.DateTime) -> List[dict]:
    """List all available CrossrefFundref releases between the start and end date

    :param start_date: The start date of the period to look for releases
    :param end_date: The end date of the period to look for releases
    :return: list with dictionaries of release info (url and release date)
    """

    # A selection of headers to prevent 403/forbidden error.
    headers_list = [
        {
            "authority": "gitlab.com",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) "
            "Chrome/84.0.4147.89 Safari/537.36",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,"
            "*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "sec-fetch-site": "none",
            "sec-fetch-mode": "navigate",
            "sec-fetch-dest": "document",
            "accept-language": "en-GB,en-US;q=0.9,en;q=0.8",
        },
        {
            "User-Agent": "Mozilla/5.0 (X11; Ubuntu; Linux x86_64; rv:78.0) Gecko/20100101 Firefox/78.0",
            "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,*/*;q=0.8",
            "Accept-Language": "en-US,en;q=0.5",
            "DNT": "1",
            "Connection": "keep-alive",
            "Upgrade-Insecure-Requests": "1",
        },
    ]

    release_info = []
    headers = random.choice(headers_list)
    current_page = 1

    while True:
        # Fetch page
        url = f"{CrossrefFundrefTelescope.RELEASES_URL}?per_page=100&page={current_page}"
        response = retry_session().get(url, headers=headers)

        # Check if correct response code
        if response is not None and response.status_code == 200:
            # Parse json
            num_pages = int(response.headers["X-Total-Pages"])
            json_response = json.loads(response.text)

            # Parse release information
            for release in json_response:
                version = float(release["tag_name"].strip("v"))
                for source in release["assets"]["sources"]:
                    if source["format"] == "tar.gz":
                        # Parse release date
                        if version == 0.1:
                            release_date = pendulum.datetime(year=2014, month=3, day=1)
                        elif version < 1.0:
                            date_string = release["description"].split("\n")[0]
                            release_date = pendulum.from_format("01 " + date_string, "DD MMMM YYYY")
                        else:
                            release_date = pendulum.parse(release["released_at"])

                        # Only include release if it is within start and end dates
                        if start_date <= release_date < end_date:
                            release_info.append({"url": source["url"], "date": release_date.format("YYYYMMDD")})

            # Check if we should exit or get the next page
            if num_pages <= current_page:
                break
            current_page += 1
        else:
            raise AirflowException(f"Error retrieving response from: {url}")

    return release_info


def strip_whitespace(file_path: str):
    """Strip leading white space from the first line of the file.
    This is present in fundref release 2019-06-01. If not removed it will give a XML ParseError.

    :param file_path: Path to file from which to trim leading white space.
    :return: None.
    """
    with open(file_path, "r") as f_in, open(file_path + ".tmp", "w") as f_out:
        first_line = True
        for line in f_in:
            if first_line and not line.startswith(" "):
                os.remove(file_path + ".tmp")
                return
            elif first_line and line.startswith(" "):
                line = line.lstrip()
            f_out.write(line)
            first_line = False
    os.rename(file_path + ".tmp", file_path)


def new_funder_template():
    """Helper Function for creating a new Funder.

    :return: a blank funder object.
    """
    return {
        "funder": None,
        "pre_label": None,
        "alt_label": [],
        "narrower": [],
        "broader": [],
        "modified": None,
        "created": None,
        "funding_body_type": None,
        "funding_body_sub_type": None,
        "region": None,
        "country": None,
        "country_code": None,
        "state": None,
        "tax_id": None,
        "continuation_of": [],
        "renamed_as": [],
        "replaces": [],
        "affil_with": [],
        "merged_with": [],
        "incorporated_into": [],
        "is_replaced_by": [],
        "incorporates": [],
        "split_into": [],
        "status": None,
        "merger_of": [],
        "split_from": None,
        "formly_known_as": None,
        "notation": None,
    }


def parse_fundref_registry_rdf(registry_file_path: str) -> Tuple[List, Dict]:
    """Helper function to parse a fundref registry rdf file and to return a python list containing each funder.

    :param registry_file_path: the filename of the registry.rdf file to be parsed.
    :return: funders list containing all the funders parsed from the input rdf and dictionary of funders with their
    id as key.
    """

    funders = []
    funders_by_key = {}

    tree = ET.parse(registry_file_path)
    root = tree.getroot()
    tag_prefix = root.tag.split("}")[0] + "}"
    for record in root:
        tag = record.tag.split("}")[-1]
        if tag == "ConceptScheme":
            for nested in record:
                tag = nested.tag.split("}")[-1]
                if tag == "hasTopConcept":
                    funder_id = nested.attrib[tag_prefix + "resource"]
                    funders_by_key[funder_id] = new_funder_template()

        if tag == "Concept":
            funder_id = record.attrib[tag_prefix + "about"]
            funder = funders_by_key[funder_id]
            funder["funder"] = funder_id
            for nested in record:
                tag = nested.tag.split("}")[-1]
                if tag == "inScheme":
                    continue
                elif tag == "prefLabel":
                    funder["pre_label"] = nested[0][0].text
                elif tag == "altLabel":
                    alt_label = nested[0][0].text
                    if alt_label is not None:
                        funder["alt_label"].append(alt_label)
                elif tag == "narrower":
                    funder["narrower"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "broader":
                    funder["broader"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "modified":
                    funder["modified"] = nested.text
                elif tag == "created":
                    funder["created"] = nested.text
                elif tag == "fundingBodySubType":
                    funder["funding_body_type"] = nested.text
                elif tag == "fundingBodyType":
                    funder["funding_body_sub_type"] = nested.text
                elif tag == "region":
                    funder["region"] = nested.text
                elif tag == "country":
                    funder["country"] = nested.text
                elif tag == "state":
                    funder["state"] = nested.text
                elif tag == "address":
                    funder["country_code"] = nested[0][0].text
                elif tag == "taxId":
                    funder["tax_id"] = nested.text
                elif tag == "continuationOf":
                    funder["continuation_of"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "renamedAs":
                    funder["renamed_as"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "replaces":
                    funder["replaces"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "affilWith":
                    funder["affil_with"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "mergedWith":
                    funder["merged_with"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "incorporatedInto":
                    funder["incorporated_into"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "isReplacedBy":
                    funder["is_replaced_by"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "incorporates":
                    funder["incorporates"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "splitInto":
                    funder["split_into"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "status":
                    funder["status"] = nested.attrib[tag_prefix + "resource"]
                elif tag == "mergerOf":
                    funder["merger_of"].append(nested.attrib[tag_prefix + "resource"])
                elif tag == "splitFrom":
                    funder["split_from"] = nested.attrib[tag_prefix + "resource"]
                elif tag == "formerlyKnownAs":
                    funder["formly_known_as"] = nested.attrib[tag_prefix + "resource"]
                elif tag == "notation":
                    funder["notation"] = nested.text
                else:
                    logging.info(f"Unrecognized tag for element: {nested}")

            funders.append(funder)

    return funders, funders_by_key


def add_funders_relationships(funders: List, funders_by_key: Dict) -> List:
    """Adds any children/parent relationships to funder instances in the funders list.

    :param funders: List of funders
    :param funders_by_key: Dictionary of funders with their id as key.
    :return: funders with added relationships.
    """

    for funder in funders:
        children, returned_depth = recursive_funders(funders_by_key, funder, 0, "narrower", [])
        funder["children"] = children
        funder["bottom"] = len(children) > 0

        parent, returned_depth = recursive_funders(funders_by_key, funder, 0, "broader", [])
        funder["parents"] = parent
        funder["top"] = len(parent) > 0

    return funders


def recursive_funders(
    funders_by_key: Dict, funder: Dict, depth: int, direction: str, sub_funders: List
) -> Tuple[List, int]:
    """Recursively goes through a funder/sub_funder dict. The funder properties can be looked up with the
    funders_by_key dictionary that stores the properties per funder id. Any children/parents for the funder are
    already given in the xml element with the 'narrower' and 'broader' tags. For each funder in the list,
    it will recursively add any children/parents for those funders in 'narrower'/'broader' and their funder properties.

    :param funders_by_key: dictionary with id as key and funders object as value
    :param funder: dictionary of a given funder containing 'narrower' and 'broader' info
    :param depth: keeping track of nested depth
    :param direction: either 'narrower' or 'broader' to get 'children' or 'parents'
    :param sub_funders: list to keep track of which funder ids are parents
    :return: list of children and current depth
    """

    starting_depth = depth
    children = []
    # Loop through funder_ids in 'narrower' or 'broader' info
    for funder_id in funder[direction]:
        if funder_id in sub_funders:
            # Stop recursion if funder is it's own parent or child
            logging.info(f"Funder {funder_id} is it's own parent/child, skipping..")
            name = "NA"
            returned = []
            returned_depth = depth
            sub_funders.append(funder_id)
        else:
            try:
                sub_funder = funders_by_key[funder_id]
                # Add funder id of sub_funder to list to keep track of 'higher' sub_funders in the recursion
                sub_funders.append(sub_funder["funder"])
                # Store name to pass on to child object
                name = sub_funder["pre_label"]
                # Get children/parents of sub_funder
                returned, returned_depth = recursive_funders(
                    funders_by_key, sub_funder, starting_depth + 1, direction, sub_funders
                )
            except KeyError:
                logging.info(f"Could not find funder by id: {funder_id}, skipping..")
                name = "NA"
                returned = []
                returned_depth = depth
                sub_funders.append(funder_id)

        # Add child/parent (containing nested children/parents) to list
        if direction == "narrower":
            child = {"funder": funder_id, "name": name, "children": returned}
        else:
            child = {"funder": funder_id, "name": name, "parent": returned}
        children.append(child)
        sub_funders.pop(-1)
        if returned_depth > depth:
            depth = returned_depth
    return children, depth



# Copyright 2020 Curtin University
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Author: Aniek Roelofs, James Diprose

import os

import httpretty
import pendulum
import vcr

from academic_observatory_workflows.config import test_fixtures_folder
from academic_observatory_workflows.workflows.geonames_telescope import (
    GeonamesRelease,
    GeonamesTelescope,
    fetch_release_date,
    first_sunday_of_month,
)
from observatory.platform.utils.file_utils import _hash_file
from observatory.platform.utils.gc_utils import bigquery_sharded_table_id
from observatory.platform.utils.test_utils import (
    ObservatoryEnvironment,
    ObservatoryTestCase,
    module_file_path,
)
from observatory.platform.utils.workflow_utils import (
    SubFolder,
    blob_name,
    workflow_path,
)


class TestGeonamesTelescope(ObservatoryTestCase):
    """Tests for the Geonames telescope"""

    def __init__(self, *args, **kwargs):
        """Constructor which sets up variables used by tests.

        :param args: arguments.
        :param kwargs: keyword arguments.
        """

        super(TestGeonamesTelescope, self).__init__(*args, **kwargs)
        self.project_id = os.getenv("TEST_GCP_PROJECT_ID")
        self.data_location = os.getenv("TEST_GCP_DATA_LOCATION")
        self.all_countries_path = test_fixtures_folder("geonames", "allCountries.zip")
        self.fetch_release_date_path = test_fixtures_folder("geonames", "fetch_release_date.yaml")
        self.list_releases_path = test_fixtures_folder("geonames", "list_releases.yaml")

    def test_dag_structure(self):
        """Test that the Geonames DAG has the correct structure.

        :return: None
        """

        dag = GeonamesTelescope().make_dag()
        self.assert_dag_structure(
            {
                "check_dependencies": ["fetch_release_date"],
                "fetch_release_date": ["download"],
                "download": ["upload_downloaded"],
                "upload_downloaded": ["extract"],
                "extract": ["transform"],
                "transform": ["upload_transformed"],
                "upload_transformed": ["bq_load"],
                "bq_load": ["cleanup"],
                "cleanup": [],
            },
            dag,
        )

    def test_dag_load(self):
        """Test that the Geonames DAG can be loaded from a DAG bag.

        :return: None
        """

        with ObservatoryEnvironment().create():
            dag_file = os.path.join(module_file_path("academic_observatory_workflows.dags"), "geonames_telescope.py")
            self.assert_dag_load("geonames", dag_file)

    def test_first_sunday_of_month(self):
        """Test first_sunday_of_month function.

        :return: None.
        """

        # Test when the date is later in the month
        datetime = pendulum.datetime(year=2020, month=7, day=28)
        expected_datetime = pendulum.datetime(year=2020, month=7, day=5)
        actual_datetime = first_sunday_of_month(datetime)
        self.assertEqual(expected_datetime, actual_datetime)

        # Test a date when the current date is a Sunday
        datetime = pendulum.datetime(year=2020, month=11, day=1)
        expected_datetime = pendulum.datetime(year=2020, month=11, day=1)
        actual_datetime = first_sunday_of_month(datetime)
        self.assertEqual(expected_datetime, actual_datetime)

    def test_fetch_release_date(self):
        """Test fetch_release_date function.

        :return: None.
        """

        with vcr.use_cassette(self.fetch_release_date_path):
            date = fetch_release_date()
            self.assertEqual(date, pendulum.datetime(year=2020, month=7, day=16, hour=1, minute=22, second=15))

    def test_telescope(self):
        """Test the Geonames telescope end to end.

        :return: None.
        """

        # Setup Observatory environment
        env = ObservatoryEnvironment(self.project_id, self.data_location)
        dataset_id = env.add_dataset()

        # Setup Telescope
        execution_date = pendulum.datetime(year=2020, month=11, day=1)
        telescope = GeonamesTelescope(dataset_id=dataset_id)
        dag = telescope.make_dag()

        # Create the Observatory environment and run tests
        with env.create():
            with env.create_dag_run(dag, execution_date):
                # Release settings
                release_date = pendulum.datetime(year=2021, month=3, day=5, hour=1, minute=34, second=32)
                release_id = f'{telescope.dag_id}_{release_date.strftime("%Y_%m_%d")}'
                download_folder = workflow_path(SubFolder.downloaded, telescope.dag_id, release_id)
                extract_folder = workflow_path(SubFolder.extracted, telescope.dag_id, release_id)
                transform_folder = workflow_path(SubFolder.transformed, telescope.dag_id, release_id)

                # Test that all dependencies are specified: no error should be thrown
                env.run_task(telescope.check_dependencies.__name__, dag, execution_date)

                # Test list releases task
                with vcr.use_cassette(self.list_releases_path):
                    ti = env.run_task(telescope.fetch_release_date.__name__, dag, execution_date)

                pulled_release_date = ti.xcom_pull(
                    key=GeonamesTelescope.RELEASE_INFO,
                    task_ids=telescope.fetch_release_date.__name__,
                    include_prior_dates=False,
                )
                self.assertIsInstance(pendulum.parse(pulled_release_date), pendulum.DateTime)
                self.assertEqual(release_date.date(), pendulum.parse(pulled_release_date).date())

                # Test download task
                with httpretty.enabled():
                    self.setup_mock_file_download(GeonamesRelease.DOWNLOAD_URL, self.all_countries_path)
                    env.run_task(telescope.download.__name__, dag, execution_date)

                download_file_path = os.path.join(download_folder, f"{telescope.dag_id}.zip")
                expected_file_hash = _hash_file(self.all_countries_path, algorithm="md5")
                self.assert_file_integrity(download_file_path, expected_file_hash, "md5")

                # Test that file uploaded
                env.run_task(telescope.upload_downloaded.__name__, dag, execution_date)
                self.assert_blob_integrity(env.download_bucket, blob_name(download_file_path), download_file_path)

                # Test that file extracted
                env.run_task(telescope.extract.__name__, dag, execution_date)
                extracted_file_path = os.path.join(extract_folder, "allCountries.txt")
                expected_file_hash = "de1bf005df4840d16faf598999d72051"
                self.assert_file_integrity(extracted_file_path, expected_file_hash, "md5")

                # Test that file transformed
                env.run_task(telescope.transform.__name__, dag, execution_date)
                transformed_file_path = os.path.join(transform_folder, f"{telescope.dag_id}.csv.gz")
                expected_file_hash = "26c14e16"
                self.assert_file_integrity(transformed_file_path, expected_file_hash, "gzip_crc")

                # Test that transformed file uploaded
                env.run_task(telescope.upload_transformed.__name__, dag, execution_date)
                self.assert_blob_integrity(
                    env.transform_bucket, blob_name(transformed_file_path), transformed_file_path
                )

                # Test that data loaded into BigQuery
                env.run_task(telescope.bq_load.__name__, dag, execution_date)
                table_id = f"{self.project_id}.{dataset_id}.{bigquery_sharded_table_id(telescope.dag_id, release_date)}"
                expected_rows = 50
                self.assert_table_integrity(table_id, expected_rows)

                # Test that all telescope data deleted
                env.run_task(telescope.cleanup.__name__, dag, execution_date)
                self.assert_cleanup(download_folder, extract_folder, transform_folder)
